package gi;

import gi.ExpressionContextMatrix.Context;
import gi.ExpressionContextMatrix.Expression;
import gi.Sentence.Element;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A bi-cluster of a SymbolSymbolMatrix. Notice that this class can represent a
 * partial bi-cluster, i.e., not all the appearances in the full bi-cluster
 * (defined by the row-header/column-header set) are included in the row/col
 * sums, {@code sum} and {@code ect}. This is useful when dealing with
 * bi-cluster overlapping.
 * 
 * @author Kewei Tu
 */
public class BiCluster {
	// /**
	// * The SymbolSymbolMatrix on which the bi-cluster is based
	// */
	// protected SymbolSymbolMatrix mainMatrix;

	// /**
	// * The row/column set for the bi-cluster
	// */
	// protected HashSet<Symbol> rowHeaders, colHeaders;

	/**
	 * The row/column set for the bi-cluster, mapped to the corresponding
	 * row/column sum.
	 */
	protected HashMap<Symbol, int[]> rows, cols;

	/**
	 * The sum of all the bi-cluster elements
	 */
	protected int sum;

	/**
	 * A temporary variable for the symbol-symbol matrix corresponding to the
	 * bi-cluster. Don't read it without knowing what's inside.
	 */
	// protected HashMap<Symbol, HashMap<Symbol, Collection<Element>>> bcssm;
	/**
	 * The expression-context matrix of this bi-cluster
	 */
	protected ExpressionContextMatrix ect = null;

	/**
	 * The log posterior gain by extracting this bi-cluster to an AndNode
	 */
	protected double posteriorGain;

	public BiCluster() {
		// this.mainMatrix = mainMatrix;
		rows = new HashMap<Symbol, int[]>(GrammarLearning.biClusterMaxSize * 2);
		cols = new HashMap<Symbol, int[]>(GrammarLearning.biClusterMaxSize * 2);
		sum = 0;
	}

	/**
	 * Get the row-headers of the bi-cluster, along with their appearance
	 * numbers in the bi-cluster.
	 * 
	 * @param rowHeaders
	 *            an empty list where the row-headers are to be stored
	 * @param appNums
	 *            an empty list where the row-header appearance numbers are to
	 *            be stored
	 */
	public void getRows(List<Symbol> rowHeaders, List<double[]> appNums) {
		for (Map.Entry<Symbol, int[]> e : rows.entrySet()) {
			rowHeaders.add(e.getKey());
			int[] a = e.getValue();
			double[] tmp = new double[a.length];
			for (int i = 0; i < a.length; i++) {
				tmp[i] = a[i];
			}
			appNums.add(tmp);
		}
	}

	/**
	 * Get the col-headers of the bi-cluster, along with their appearance
	 * numbers in the bi-cluster.
	 * 
	 * @param colHeaders
	 *            an empty list where the col-headers are to be stored
	 * @param appNums
	 *            an empty list where the col-header appearance numbers are to
	 *            be stored
	 */
	public void getCols(List<Symbol> colHeaders, List<double[]> appNums) {
		for (Map.Entry<Symbol, int[]> e : cols.entrySet()) {
			colHeaders.add(e.getKey());
			int[] a = e.getValue();
			double[] tmp = new double[a.length];
			for (int i = 0; i < a.length; i++) {
				tmp[i] = a[i];
			}
			appNums.add(tmp);
		}
	}

	@Override
	public String toString() {
		String re = "Rows: ";
		for (Symbol s : rows.keySet())
			re += s + ", ";
		re += "\nCols: ";
		for (Symbol s : cols.keySet())
			re += s + ", ";

		if (ect != null) {
			re += "\n\n";
			final int sizelimit = 50;
			ArrayList<Symbol> cs = new ArrayList<Symbol>(sizelimit);
			int count = 0;
			for (Symbol c : cols.keySet()) {
				cs.add(c);
				count++;
				if (count == sizelimit)
					break;
			}
			for (Object ch : cs)
				re += "\t" + ch;
			if (cols.size() > sizelimit)
				re += "\t(" + (cols.size() - sizelimit) + " more cols omitted)";
			count = 0;
			for (Symbol r : rows.keySet()) {
				re += "\n" + r;
				for (Symbol ch : cs) {
					re += "\t";
					Map<Context, Collection<Element>> ectrow = ect.rows
							.get(new Expression(r, ch));
					if (ectrow != null) {
						int size = 0;
						for (Collection<Element> cell : ectrow.values())
							size += cell.size();
						re += size;
					}
				}
				count++;
				if (count == sizelimit)
					break;
			}
			if (rows.size() > sizelimit)
				re += "\n(" + (rows.size() - sizelimit) + " more rows omitted)";
		}

		return re;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof BiCluster) {
			BiCluster bc2 = (BiCluster) obj;
			return rows.keySet().equals(bc2.rows.keySet())
					&& cols.keySet().equals(bc2.cols.keySet());
		} else
			return false;
	}
}
