package gi;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Scanner;

public class Corpus {
	protected Sentence[] sentences;
	protected Term terms[];

	public Corpus(Sentence[] sentences, Term[] terms) {
		this.sentences = sentences;
		this.terms = terms;
	}

	/**
	 * Import a pre-processed corpus. The file format is as follows:
	 * <li> Term list: the first line is the number of terms; then the list of
	 * all the individual terms in the corpus, one term at each line.
	 * <li> A blank line.
	 * <li> Sentence list: the first line is the number of sentences; then the
	 * list of all the sentences, one at each line. Each sentence is a sequence
	 * of terms represented by their (0-based) line numbers in the term list.
	 * 
	 * @param filename
	 *            the name of the file containing the pre-processed corpus
	 */
	public Corpus(String filename) {
		Scanner s = null;
		try {
			// read in terms
			s = new Scanner(new BufferedReader(new FileReader(filename)));
			int len = 0;
			if (s.hasNext())
				len = s.nextInt();
			terms = new Term[len];
			s.nextLine();
			for (int i = 0; i < len; i++) {
				terms[i] = new Term(s.nextLine());
			}

			s.nextLine();

			// read in sentences
			len = 0;
			if (s.hasNext())
				len = s.nextInt();
			sentences = new Sentence[len];
			s.nextLine(); // go to the 2nd line
			for (int i = 0; i < len; i++) {
				Scanner ss = new Scanner(s.nextLine());
				Sentence.Element first = null, prev = null;
				if (ss.hasNext()) {
					Term t = terms[ss.nextInt()];
					t.appearanceNum++;
					first = prev = new Sentence.Element(t);
					while (ss.hasNext()) {
						t = terms[ss.nextInt()];
						t.appearanceNum++;
						Sentence.Element e = new Sentence.Element(t);
						e.prev = prev;
						prev.next = e;
						prev = e;
					}
				}
				sentences[i] = new Sentence(first);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} finally {
			if (s != null) {
				s.close();
			}
		}
	}

	public void saveToFile(String filename) {
		try {
			// store terms
			FileWriter fw = new FileWriter(filename);
			fw.write(terms.length + "\n");
			for (int i = 0; i < terms.length; i++) {
				fw.write(terms[i] + "\n");
			}
			// fw.close();

			fw.write("\n");

			HashMap<Term, Integer> termMap = new HashMap<Term, Integer>(
					(int) (terms.length / 0.75) + 1);
			for (int i = 0; i < terms.length; i++) {
				termMap.put(terms[i], i);
			}

			// store sentences
			// fw = new FileWriter(filenames[1]);
			fw.write(sentences.length + "\n");
			for (int i = 0; i < sentences.length; i++) {
				for (Sentence.Element j = sentences[i].firstElem; j != null; j = j.next) {
					fw.write(termMap.get(j.sym) + " ");
				}
				fw.write("\n");
			}

			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void saveToReadableFile(String filename) {
		try {
			FileWriter fw = new FileWriter(filename);

			for (int i = 0; i < sentences.length; i++) {
				for (Sentence.Element j = sentences[i].firstElem; j != null; j = j.next) {
					fw.write(j.sym + " ");
				}
				fw.write("\n");
			}

			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/*
	 * The following methods are used for cross-validation
	 */
	/**
	 * Split the corpus to n corpora of (roughly) equal size. Notice that the
	 * new Corpus objects share the terms and sentences with the original Corpus
	 * object.
	 * 
	 * @param n
	 * @return
	 */
	protected Corpus[] split(int n) {
		Corpus[] re = new Corpus[n];
		int size = sentences.length / n;
		for (int i = 0; i < n - 1; i++) {
			re[i] = new Corpus(Arrays.copyOfRange(sentences, i * size, (i + 1)
					* size), terms);
		}
		re[n - 1] = new Corpus(Arrays.copyOfRange(sentences, (n - 1) * size,
				sentences.length), terms);
		return re;
	}

	/**
	 * Merge this Corpus and the input Corpus into one. The two Corpus to be
	 * merged must have the same value for the {@code terms} field. Notice that
	 * the new Corpus object shares the terms and sentences with the original
	 * two Corpus objects.
	 * 
	 * @param c2
	 * @return
	 */
	protected Corpus merge(Corpus c2) {
		if (terms != c2.terms)
			return null;
		Sentence[] sent = Arrays.copyOf(sentences, sentences.length
				+ c2.sentences.length);
		for (int i = sentences.length; i < sent.length; i++) {
			sent[i] = c2.sentences[i - sentences.length];
		}
		return new Corpus(sent, terms);
	}

	/**
	 * Make a copy of the Corpus, where all the sentences and terms are
	 * duplicated.
	 * 
	 * @return
	 */
	protected Corpus copy() {
		return copyExceptPart(0, 0);
	}

	/**
	 * Make a copy of a part of the Corpus, where the sentences outside the
	 * specified index range are duplicated and all the terms are duplicated.
	 * 
	 * @param from
	 *            inclusive start index.
	 * @param to
	 *            exclusive end index.
	 * @return
	 */
	protected Corpus copyExceptPart(int from, int to) {
		if (from < 0)
			from = 0;
		if (to > sentences.length)
			to = sentences.length;

		Term[] t = new Term[terms.length];
		HashMap<Term, Term> termMap = new HashMap<Term, Term>(
				(int) (terms.length / 0.75) + 1);
		for (int i = 0; i < t.length; i++) {
			t[i] = new Term(terms[i].obj);
			termMap.put(terms[i], t[i]);
		}
		Sentence[] sent = new Sentence[sentences.length - (to - from)];
		for (int i = 0; i < from; i++) {
			if (sentences[i].firstElem == null)
				sent[i] = new Sentence(null);
			else {
				Symbol firstSym = termMap.get(sentences[i].firstElem.sym);
				firstSym.appearanceNum++;
				Sentence.Element first = new Sentence.Element(firstSym);
				Sentence.Element p2 = first;
				for (Sentence.Element p = sentences[i].firstElem.next; p != null; p = p.next) {
					Symbol sym = termMap.get(p.sym);
					sym.appearanceNum++;
					p2.next = new Sentence.Element(sym);
					p2.next.prev = p2;
					p2 = p2.next;
				}
				sent[i] = new Sentence(first);
			}
		}
		for (int i = to; i < sentences.length; i++) {
			if (sentences[i].firstElem == null)
				sent[i - to + from] = new Sentence(null);
			else {
				Symbol firstSym = termMap.get(sentences[i].firstElem.sym);
				firstSym.appearanceNum++;
				Sentence.Element first = new Sentence.Element(firstSym);
				Sentence.Element p2 = first;
				for (Sentence.Element p = sentences[i].firstElem.next; p != null; p = p.next) {
					Symbol sym = termMap.get(p.sym);
					sym.appearanceNum++;
					p2.next = new Sentence.Element(sym);
					p2.next.prev = p2;
					p2 = p2.next;
				}
				sent[i - to + from] = new Sentence(first);
			}
		}
		return new Corpus(sent, t);
	}

}
