package gi;

import gi.Sentence.Element;
import gi.SymbolSymbolMatrix.SymbolSymbolCell;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

/**
 * An expression-context matrix. An expression is a pair of symbols. A context
 * is the symbol either directly before the pair or directly after the pair; it
 * is null if there's no symbol before/after the pair, i.e., the pair is the
 * beginning/end of the sentence. Notice that for each {@link Sentence.Element}
 * contained in the matrix, it appears exactly twice, one for the left context
 * and one for the right.
 * 
 * @author Kewei Tu
 * 
 */
public class ExpressionContextMatrix
		extends
		Matrix<ExpressionContextMatrix.Expression, ExpressionContextMatrix.Context, Collection<Element>> {

	static public class Expression {
		protected Symbol s1, s2;

		public Expression(Symbol s1, Symbol s2) {
			this.s1 = s1;
			this.s2 = s2;
		}

		@Override
		public boolean equals(Object obj) {
			Expression b = (Expression) obj;
			return (s1 == b.s1) && (s2 == b.s2);
		}

		@Override
		public int hashCode() {
			return s1.hashCode() + s2.hashCode();
		}

		@Override
		public String toString() {
			return "(" + s1 + "," + s2 + ")";
		}
	}

	static public class Context {
		/**
		 * {@code sym} can be null, which represents either the beginning or the
		 * end of a sentence, depending on {@code contextType}.
		 */
		protected Symbol sym;

		protected int contextType;

		public Context(Symbol sym, int contextType) {
			this.sym = sym;
			this.contextType = contextType;
		}

		public static final int LEFT_NEIGHBOR = 0;
		public static final int RIGHT_NEIGHBOR = 1;

		@Override
		public boolean equals(Object obj) {
			Context b = (Context) obj;
			return (sym == null ? b.sym == null : sym == b.sym)
					&& (contextType == b.contextType);
		}

		@Override
		public int hashCode() {
			return (sym == null ? 0 : sym.hashCode()) + contextType;
		}

		@Override
		public String toString() {
			return sym + "("
					+ (contextType == LEFT_NEIGHBOR ? "Left" : "Right") + ")";
		}
	}

//	/**
//	 * @param s1
//	 *            the 1st symbol of the pair
//	 * @param s2
//	 *            the 2nd symbol of the pair
//	 * @param apps
//	 *            the appearances of the pair
//	 */
//	public void addRow(Symbol s1, Symbol s2, Collection<Sentence.Element> apps) {
//		Expression exp = new Expression(s1, s2);
//		addRow(exp);
//		Map<Context, Collection<Sentence.Element>> row = rows.get(exp);
//		for (Sentence.Element elem : apps) {
//			assert elem.sym == s1;
//			assert elem.next.sym == s2;
//
//			// left
//			Symbol s = elem.prev != null ? elem.prev.sym : null;
//			Context cont = new Context(s, Context.LEFT_NEIGHBOR);
//			if (cols.get(cont) == null)
//				addCol(cont);
//			if (row.get(cont) == null) {
//				Collection<Sentence.Element> c = new ArrayList<Sentence.Element>();
//				c.add(elem);
//				row.put(cont, c);
//				cols.get(cont).put(exp, c);
//			} else {
//				row.get(cont).add(elem);
//			}
//
//			// right
//			s = elem.next.next != null ? elem.next.next.sym : null;
//			cont = new Context(s, Context.RIGHT_NEIGHBOR);
//			if (cols.get(cont) == null)
//				addCol(cont);
//			if (row.get(cont) == null) {
//				Collection<Sentence.Element> c = new ArrayList<Sentence.Element>();
//				c.add(elem);
//				row.put(cont, c);
//				cols.get(cont).put(exp, c);
//			} else {
//				row.get(cont).add(elem);
//			}
//		}
//		if (row.size() == 0)
//			rows.remove(exp);
//	}

	/**
	 * @param elem
	 *            the first element of the pair to be deleted
	 * @return whether the pair is found and deleted
	 */
	public boolean delAppearance(Sentence.Element elem) {
		Expression pair = new Expression(elem.sym, elem.next.sym);
		Map<Context, Collection<Sentence.Element>> row = rows.get(pair);
		if (row != null) {
			boolean found = false;
			Symbol s = elem.prev != null ? elem.prev.sym : null;
			Context cont = new Context(s, Context.LEFT_NEIGHBOR);
			Collection<Sentence.Element> c = row.get(cont);
			if (c != null)
				if (c.remove(elem)) {
					found = true;
					if (c.size() == 0)
						delCell(pair, cont);
				}
			if (found) {
				s = elem.next.next != null ? elem.next.next.sym : null;
				cont = new Context(s, Context.RIGHT_NEIGHBOR);
				c = row.get(cont);
				if (c.remove(elem)) {
					if (c.size() == 0)
						delCell(pair, cont);
				}
			}
			return found;
		}
		return false;
	}

	@Override
	protected String cellToString(Collection<Element> c) {
		return Integer.toString(c.size());
	}

	public void delAppearance(Expression row, Context col, Element app) {
		Map<Context, Collection<Sentence.Element>> r = rows.get(row);
		if (r != null) {
			Collection<Sentence.Element> apps = r.get(col);
			if (apps != null)
				if (apps.remove(app))
					if (apps.isEmpty())
						delCell(row, col);
		}
	}

	/**
	 * Add a row according to cell.ecv. Ecv is only for reading and won't be
	 * incorporated into ECT.
	 * 
	 * @param s1
	 * @param s2
	 * @param cell
	 */
	public void addRow(Symbol s1, Symbol s2, SymbolSymbolCell cell) {
		Map<Context, Collection<Element>> ecv = cell.ecv;
		Expression exp = new Expression(s1, s2);
		/*
		 * TODO v3: In v3, the elements in a Collection won't be separated in
		 * the life of this ECT. So we can directly reuse the Collection
		 * objects.
		 */
		// Map<Context, Collection<Sentence.Element>> row = new HashMap<Context,
		// Collection<Element>>(ecv);
		Map<Context, Collection<Sentence.Element>> row = new HashMap<Context, Collection<Element>>();
		for (Map.Entry<Context, Collection<Sentence.Element>> e : ecv
				.entrySet()) {
			row.put(e.getKey(), new HashSet<Element>(e.getValue()));
		}
		rows.put(exp, row);
		for (Map.Entry<Context, Collection<Sentence.Element>> e : row
				.entrySet()) {
			Context cont = e.getKey();
			if (cols.get(cont) == null)
				addCol(cont);
			cols.get(cont).put(exp, e.getValue());
		}
	}
}
