package gi;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Scanner;

public class Grammar {
	public ArrayList<AndNode> andNodes;

	public Term[] terms;
	protected Map<Object, Term> termMap;

	public Term getTerminal(Object o) {
		return termMap.get(o);
	}

	public ArrayList<OrNode> orNodes; // including the start symbol

	public OrNode startSymbol;

	/*
	 * For an ideal grammar, from the start symbol s we should be able to reach
	 * all the AND nodes. This is not always true, although those unreachable
	 * AND nodes won't be used either in parsing or in generation.
	 */

	/**
	 * Create a grammar with only terminals.
	 * 
	 * @param terms
	 */
	public Grammar(Term[] terms) {
		this.terms = terms;
		termMap = new HashMap<Object, Term>((int) (terms.length / 0.75) + 1);
		for (int i = 0; i < terms.length; i++) {
			Term term = terms[i];
			termMap.put(term.obj, term);
		}
		andNodes = new ArrayList<AndNode>();
		orNodes = new ArrayList<OrNode>();
	}

	/**
	 * Create a grammar by reading from a file.
	 * 
	 * @param filename
	 */
	public Grammar(String filename) {
		readFromFile(filename);
	}

	public void saveToFile(String filename) {
		HashMap<OrNode, Integer> orMap = new HashMap<OrNode, Integer>(
				(int) (orNodes.size() / 0.75) + 1);
		int ind = 0;
		for (OrNode or : orNodes) {
			orMap.put(or, ind);
			ind++;
		}

		HashMap<Symbol, Integer> symMap = new HashMap<Symbol, Integer>(
				(int) ((andNodes.size() + terms.length) / 0.75) + 1);
		ind = 0;
		for (int i = 0; i < terms.length; i++) {
			symMap.put(terms[i], ind);
			ind++;
		}
		for (AndNode and : andNodes) {
			symMap.put(and, ind);
			ind++;
		}

		try {
			FileWriter fw = new FileWriter(filename);
			fw.write("Term# " + terms.length + "\n");
			fw.write("AndNode# " + andNodes.size() + "\n");
			fw.write("OrNode# " + orNodes.size() + "\n");
			fw.write("StartSymbol " + orMap.get(startSymbol) + "\n");

			fw.write("Terms\n");
			for (int i = 0; i < terms.length; i++) {
				fw.write(terms[i].appearanceNum + " " + terms[i].obj + "\n");
			}

			fw.write("AndNodes\n");
			for (AndNode and : andNodes) {
				fw.write(and.appearanceNum + " " + orMap.get(and.left) + " "
						+ orMap.get(and.right) + "\n");
			}

			fw.write("OrNodes\n");
			for (OrNode or : orNodes) {
				for (int i = 0; i < or.children.size(); i++) {
					fw.write(symMap.get(or.children.get(i)) + " "
							+ or.weights.get(i)[0] + " ");
				}
				fw.write("\n");
			}

			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void readFromFile(String filename) {
		try {
			Scanner s = new Scanner(
					new BufferedReader(new FileReader(filename)));

			s.next();
			terms = new Term[s.nextInt()];

			s.next();
			int nAnd = s.nextInt();
			andNodes = new ArrayList<AndNode>(nAnd);

			s.next();
			int nOr = s.nextInt();
			orNodes = new ArrayList<OrNode>(nOr);
			for (int i = 0; i < nOr; i++) {
				orNodes.add(new OrNode(new ArrayList<Symbol>(),
						new ArrayList<double[]>()));
			}

			s.next();
			startSymbol = orNodes.get(s.nextInt());

			s.next();
			for (int i = 0; i < terms.length; i++) {
				int tmp = s.nextInt();
				terms[i] = new Term(s.nextLine().substring(1));
				terms[i].appearanceNum = tmp;
			}

			s.next();
			for (int i = 0; i < nAnd; i++) {
				int tmp = s.nextInt();
				andNodes.add(new AndNode(orNodes.get(s.nextInt()), orNodes
						.get(s.nextInt()), tmp));
			}

			s.next();
			s.nextLine();
			for (OrNode or : orNodes) {
				Scanner ss = new Scanner(s.nextLine());
				while (ss.hasNext()) {
					int i = ss.nextInt();
					Symbol chld;
					if (i < terms.length)
						chld = terms[i];
					else
						chld = andNodes.get(i - terms.length);
					or.children.add(chld);
					chld.parents.add(or);
					double[] tmp = { ss.nextDouble() };
					or.weights.add(tmp);
				}
			}

			s.close();

			termMap = new HashMap<Object, Term>((int) (terms.length / 0.75) + 1);
			for (int i = 0; i < terms.length; i++) {
				Term term = terms[i];
				termMap.put(term.obj, term);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	public void saveToReadableFile(String filename) {
		HashMap<OrNode, Integer> orMap = new HashMap<OrNode, Integer>(
				(int) (orNodes.size() / 0.75) + 1);
		int ind = 0;
		for (OrNode or : orNodes) {
			orMap.put(or, ind);
			ind++;
		}

		try {
			FileWriter fw = new FileWriter(filename);
			fw.write("(\n");

			fw.write("\t(" + orMap.get(startSymbol) + " ->");
			for (int i = 0; i < startSymbol.children.size(); i++) {
				fw.write("\n\t\t(" + startSymbol.weights.get(i)[0] + " ");
				Symbol chld = startSymbol.children.get(i);
				if (chld instanceof Term) {
					Term t = (Term) chld;
					fw.write(". " + t.obj);
				} else {
					AndNode a = (AndNode) chld;
					fw.write(orMap.get(a.left) + " " + orMap.get(a.right));
				}
				fw.write(")");
			}
			fw.write(")\n");

			ind = -1;
			for (OrNode or : orNodes) {
				ind++;
				if (or == startSymbol)
					continue;
				fw.write("\t(" + ind + " ->");
				for (int i = 0; i < or.children.size(); i++) {
					fw.write("\n\t\t(" + or.weights.get(i)[0] + " ");
					Symbol chld = or.children.get(i);
					if (chld instanceof Term) {
						Term t = (Term) chld;
						fw.write(". " + t.obj);
					} else {
						AndNode a = (AndNode) chld;
						fw.write(orMap.get(a.left) + " " + orMap.get(a.right));
					}
					fw.write(")");
				}
				fw.write(")\n");
			}

			fw.write(")\n");
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void saveToReadableFile2(String filename) {
		HashMap<OrNode, Integer> orMap = new HashMap<OrNode, Integer>(
				(int) (orNodes.size() / 0.75) + 1);
		int ind = 0;
		for (OrNode or : orNodes) {
			orMap.put(or, ind);
			ind++;
		}

		try {
			FileWriter fw = new FileWriter(filename+".cfg");

			fw.write("SIGMA\n");
			for (int i = 0; i < startSymbol.children.size(); i++) {
				Symbol chld = startSymbol.children.get(i);
				if (chld instanceof Term) {
					Term t = (Term) chld;
					fw.write(t.obj.toString());
				} else {
					AndNode a = (AndNode) chld;
					fw.write(orMap.get(a.left) + " " + orMap.get(a.right));
				}
				fw.write("\n");
			}
			fw.write("\n");

			ind = -1;
			for (OrNode or : orNodes) {
				ind++;
				if (or == startSymbol)
					continue;
				fw.write(ind + "\n");
				for (int i = 0; i < or.children.size(); i++) {
					Symbol chld = or.children.get(i);
					if (chld instanceof Term) {
						Term t = (Term) chld;
						fw.write(t.obj.toString());
					} else {
						AndNode a = (AndNode) chld;
						fw.write(orMap.get(a.left) + " " + orMap.get(a.right));
					}
					fw.write("\n");
				}
				fw.write("\n");
			}

			fw.close();
			
			fw = new FileWriter(filename+".lex");
			for (Term t : terms) {
				fw.write(t.obj+" "+t.obj+"\n");
			}
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
