package gi;

import gi.ExpressionContextMatrix.Context;
import gi.ExpressionContextMatrix.Expression;
import gi.Sentence.Element;
import gi.SymbolSymbolMatrix.SymbolSymbolCell;
import gi.util.CYKParser;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

public class GrammarLearning {
	/**
	 * @param args
	 *            The first argument is the filenames for the pre-processed
	 *            corpus. See {@link Corpus#Corpus(String)}. The second is the
	 *            filename that the learned grammar is to be saved to.
	 *            (Optional: the third is the filename that the learned grammar
	 *            is to be saved to in a human readable way.)
	 */
	public static void main(String[] args) {
		Corpus corpus = new Corpus(args[0]);
		Corpus corpus2;
		if (DEBUG_MODE)
			corpus2 = corpus.copy();

		GrammarLearning gl = new GrammarLearning();
		Grammar gram = gl.grammarLearning(corpus);
		gram.saveToFile(args[1]);
		if (args.length == 3)
			gram.saveToReadableFile(args[2]);

		// Evaluation on the training set by weak generative capacity
		if (DEBUG_MODE) {
			CYKParser parser = new CYKParser();
			parser.setGrammar(gram);
			int count = 0;
			for (int j = 0; j < corpus2.sentences.length; j++) {
				System.out.print(j + ",");
				if (parser.parse(corpus2.sentences[j]))
					count++;
			}
			System.out.println("\nCorpus parsing ratio: " + ((double) count)
					/ corpus2.sentences.length);
		}
	}

	protected SymbolSymbolMatrix ssm;

	protected double totalPosteriorGain;

	/**
	 * The grammar learning algorithm.
	 * 
	 * @param c
	 *            The corpus that the grammar is learned from. Notice that the
	 *            input corpus must be clean, in that the sentences in the
	 *            corpus must contain only terms, and the terms in the corpus
	 *            have no parents. Also notice that after learning the corpus is
	 *            destroyed, in that the sentences may contain AndNode, and the
	 *            terms may have parents.
	 * @return The learned grammar.
	 */
	public Grammar grammarLearning(Corpus c) {
		totalPosteriorGain = 0;
		ConstituentFilter cf = new ConstituentFilter(constituentFilterType);
		Grammar g = new Grammar(c.terms);

		// Initialize the symbol-symbol matrix
		ssm = new SymbolSymbolMatrix(c);
		for (Map<Symbol, SymbolSymbolCell> row : ssm.rows.values().toArray(
				new Map[0])) {
			for (SymbolSymbolCell cell : row.values().toArray(
					new SymbolSymbolCell[0])) {
				if (!cf.isConstituent(cell)) {
					Element e1 = cell.ecv.values().iterator().next().iterator()
							.next();
					Element e2 = e1.next;
					ssm.delCell(e1.sym, e2.sym);
				}
			}
		}

		assert ssm.isConsistent();

		if (DEBUG_MODE)
			debug_learning_history.add(ssm.toString());

		// Iteratively learn new nodes
		BiCluster bc;
		while ((bc = bestBiClustering()) != null) {
			// Construct the two OR nodes
			ArrayList<Symbol> children = new ArrayList<Symbol>(bc.rows.size());
			ArrayList<double[]> weights = new ArrayList<double[]>(bc.rows
					.size());
			bc.getRows(children, weights);
			OrNode o1 = new OrNode(children, weights);
			children = new ArrayList<Symbol>(bc.cols.size());
			weights = new ArrayList<double[]>(bc.cols.size());
			bc.getCols(children, weights);
			OrNode o2 = new OrNode(children, weights);
			g.orNodes.add(o1);
			g.orNodes.add(o2);

			// Construct the AND node
			AndNode n = new AndNode(o1, o2, bc.sum);
			g.andNodes.add(n);

			if (DEBUG_MODE) {
				debug_learning_history.add("Best BC: " + n + "\n"
						+ bc.toString());
				debug_learning_history.add(bc.ect.toString());
			}

			// Do reduction on the extracted bi-cluster; Insert the resulting
			// new appearances to the symbol-symbol matrix.
			doReduction(bc.ect, n); /*
									 * Note: after this point, the other fields
									 * in bc may be inconsistent with bc.ect and
									 * hence should not be used.
									 */
			Map<Symbol, SymbolSymbolCell> newRow = new HashMap<Symbol, SymbolSymbolCell>();
			Map<Symbol, SymbolSymbolCell> newCol = new HashMap<Symbol, SymbolSymbolCell>();
			for (Map.Entry<Context, Map<Expression, Collection<Element>>> e : bc.ect.cols
					.entrySet()) {
				Context cont = e.getKey();
				if (cont.sym != null && cont.sym != n) {
					Map<Symbol, SymbolSymbolCell> m = (cont.contextType == Context.LEFT_NEIGHBOR) ? newCol
							: newRow;
					SymbolSymbolCell apps = new SymbolSymbolCell();
					for (Collection<Element> app : e.getValue().values()) {
						if (cont.contextType == Context.LEFT_NEIGHBOR) {
							for (Element elem : app) {
								apps.add(elem.prev);
							}
						} else
							apps.addAll(app);
					}
					if (cf.isConstituent(apps))
						m.put(cont.sym, apps);
				}
			}
			Context nc = new Context(n, Context.RIGHT_NEIGHBOR);
			if (bc.ect.cols.containsKey(nc)) {
				/*
				 * In this case, bc.ect.cols.get(nc) must only contain one
				 * entry, i.e. the one with null expression.
				 */
				Collection<Element> cNN = bc.ect.cols.get(nc).get(null);
				SymbolSymbolCell cellNN = new SymbolSymbolCell();
				cellNN.addAll(cNN);
				if (cf.isConstituent(cellNN)) {
					newRow.put(n, cellNN);
					newCol.put(n, cellNN);
				}
			} /*
				 * we don't do this in the above loop because if so then there
				 * would be two different cNN sets in newRow/newCol.
				 */
			if (!newRow.isEmpty())
				ssm.addRow(n, newRow);
			if (!newCol.isEmpty())
				ssm.addCol(n, newCol);

			assert ssm.isConsistent();

			if (DEBUG_MODE) {
				debug_learning_history.add(ssm.toString());
			}

			// Find existing OR nodes that the new AND node should merge into
			/*
			 * TODO Enhancement: Currently a greedy method is used for
			 * disambiguation (including below and doMaxReduction()), i.e., we
			 * check candidates one by one and do reduction every time a good
			 * candidate is found. A better way would be, repeatedly checking
			 * all the candidates and choosing the best one (according to the
			 * posterior) for reduction.
			 */
			for (int ii = 0; ii < 2; ii++) {
				Map<Symbol, SymbolSymbolCell> rowOrCol = ii == 0 ? newRow
						: newCol;

				HashSet<OrNode> candidates = new HashSet<OrNode>();
				for (Symbol sym : rowOrCol.keySet()) {
					for (OrNode orNode : sym.parents) {
						OrNode o = ii == 0 ? orNode.parents.get(0).right
								: orNode.parents.get(0).left;
						if (o == orNode)
							candidates.add(orNode);
					}
				}
				for (OrNode or : candidates) {
					if (DEBUG_MODE)
						debug_mg_nCandidate++;

					/* Thresholding 1 */
					double sum = 0;
					int rsum = 0;
					for (int i = 0; i < or.children.size(); i++) {
						sum += or.weights.get(i)[0];
						SymbolSymbolCell apps = rowOrCol
								.get(or.children.get(i));
						if (apps != null)
							rsum += apps.size();
					}

					if (rsum == 0)
						continue;

					double lhg_1 = 0;
					lhg_1 += xlogx(rsum);
					lhg_1 += xlogx(sum);
					lhg_1 -= xlogx(sum + rsum);
					for (int i = 0; i < or.children.size(); i++) {
						SymbolSymbolCell apps = rowOrCol
								.get(or.children.get(i));
						if (apps != null) {
							int ai = apps.size();
							double ci = or.weights.get(i)[0];
							lhg_1 -= xlogx(ci);
							lhg_1 += xlogx(ci + ai);
							lhg_1 -= xlogx(ai);
						} // else it's "+=0"
					}
					int normalizing = NORMALIZATION_MODE == 1 ? rsum
							: or.children.size();
					if (-lhg_1 > normalizing * delta)
						continue;

					if (DEBUG_MODE)
						debug_mg_nPastCheck1++;

					/* Construct ECT for N-or */
					ExpressionContextMatrix ect = new ExpressionContextMatrix();
					for (Symbol s : or.children) {
						SymbolSymbolCell cell = rowOrCol.get(s);
						if (cell != null) {
							if (ii == 0)
								ect.addRow(n, s, cell);
							else
								ect.addRow(s, n, cell);
						}
					}

					/* Construct the derived ECT for or's AND */
					/*
					 * Note: the constructed ECT here may be incomplete because
					 * 1) there's constituent-filtering loss in ssm; 2) ect0
					 * can't contain sentence beginning/end contexts. (TODO v3
					 * del #2)
					 * 
					 * For the constituent-filtering loss problem:
					 * 
					 * If the constituent-filter is very weak or doesnt exist,
					 * then its not a problem.
					 * 
					 * If the constituent-filter is strong and accurate, then we
					 * can't easily recover the original ECT. In order to make
					 * the two ECT ("ect", "ect0") comparable, we have to do
					 * constituent-filtering for "ect", as well as make a
					 * constituent-filtered version of "ect0" (i.e., not include
					 * non-constituent contexts).
					 */
					AndNode and = or.parents.get(0);
					HashMap<Context, double[]> ect0 = new HashMap<Context, double[]>();

					if (cf.isWeak()) {
						inferContext(and, ect0, true, 1, 0);
						inferContext(and, ect0, false, 1, 0);
					} else {
						/*
						 * TODO make a constituent-filtered version of "ect0"
						 * (i.e., not include non-constituent contexts)
						 */
						assert false; // a reminder
					}

					/* Overlapping filtering */
					/*
					 * TODO Enhancement: see to-do comments on overlapping
					 * filtering in BiClustering().
					 */
					if (DEBUG_MODE)
						debug_of_mg_nBeforeOF += rsum;
					HashSet<Expression> abnormals = new HashSet<Expression>();
					double mult = (rsum * mcbcMaxRatioToPrediction)
							/ (rsum + sum);
					for (int i = 0; i < or.children.size(); i++) {
						SymbolSymbolCell apps = rowOrCol
								.get(or.children.get(i));
						int ai = (apps != null) ? apps.size() : 0;
						double ci = or.weights.get(i)[0];
						if (ai > mult * (ai + ci)) { // need to check
							// overlapping
							Expression exp = ii == 0 ? new Expression(n,
									or.children.get(i)) : new Expression(
									or.children.get(i), n);
							abnormals.add(exp);
						}
					}
					for (Expression exp : abnormals) {
						Map<Context, Collection<Element>> r = ect.rows.get(exp);
						if (r != null) {
							Context[] conts = r.keySet()
									.toArray(new Context[0]);
							for (int i = 0; i < conts.length; i++) {
								Context cont = conts[i];
								if (ect0.get(cont) == null) {
									Map<Expression, Collection<Element>> column = ect.cols
											.get(cont);
									if (column != null)
										if (abnormals.containsAll(column
												.keySet())) {
											for (Map.Entry<Expression, Collection<Element>> e : column
													.entrySet()) {
												for (Element elem : e
														.getValue()) {
													boolean left = (cont.contextType == Context.LEFT_NEIGHBOR);
													Element neighbor2 = left ? elem.next.next
															: elem.prev;
													/*
													 * TODO Enhancement: Here we
													 * use an OR condition,
													 * i.e., when either side of
													 * the context meets the
													 * condition, the appearance
													 * is filtered. An
													 * alternative would be
													 * using an AND condition,
													 * which is less strict. The
													 * current one might be too
													 * strict because ect0 is
													 * very incomplete due to
													 * constituent filtering and
													 * bi-cluster extraction,
													 * making this filtering
													 * more likely.
													 */
													Symbol s = neighbor2 != null ? neighbor2.sym
															: null;
													Context cont2 = new Context(
															s,
															left ? Context.RIGHT_NEIGHBOR
																	: Context.LEFT_NEIGHBOR);
													ect.delAppearance(e
															.getKey(), cont2,
															elem);
												}
											}
											ect.delCol(cont);
										}
								}
							}
						}
					}
					if (ect.isEmpty())
						continue;

					/* Thresholding 2 */
					if (!cf.isWeak()) { // TODO v3 del
						/*
						 * TODO make a constituent-filtered version of ect, and
						 * use that version to check the coherence. See comments
						 * in "construct the derived ECT" above. Note: keep a
						 * copy of the original ect, which will be used when
						 * doing merging.
						 */
						assert false; // a reminder
					}

					lhg_1 = 0;
					double lhg_2 = 0;
					rsum = 0;
					Map<Symbol, int[]> newRowStat = new HashMap<Symbol, int[]>(
							or.children.size());
					for (int i = 0; i < or.children.size(); i++) {
						Symbol sym = or.children.get(i);
						Expression exp;
						if (ii == 0)
							exp = new Expression(n, sym);
						else
							exp = new Expression(sym, n);
						Map<Context, Collection<Element>> ectrow = ect.rows
								.get(exp);
						if (ectrow != null) {
							int ai = 0;
							for (Collection<Element> v : ectrow.values()) {
								ai += v.size();
							}
							ai /= 2; // TODO v3 del /2
							rsum += ai;
							double ci = or.weights.get(i)[0];
							lhg_1 -= xlogx(ci);
							lhg_1 += xlogx(ci + ai);
							lhg_1 -= xlogx(ai);
							lhg_2 += xlogx(ai * 2);

							int[] tmp = { ai };
							newRowStat.put(sym, tmp);
						} // else it's "+=0"
					}

					if (DEBUG_MODE)
						debug_of_mg_nAfterOF += rsum;

					if (rsum == 0)
						continue;

					lhg_1 += xlogx(rsum);
					lhg_1 += xlogx(sum);
					lhg_1 -= xlogx(sum + rsum);
					lhg_2 += xlogx(sum * 2);
					lhg_2 -= xlogx(2 * (sum + rsum));

					// calculate sentence beginning/end in ect
					// TODO v3 no need
					double left = 0, right = 0;
					for (Map.Entry<Context, double[]> e : ect0.entrySet()) {
						if (e.getKey().contextType == Context.LEFT_NEIGHBOR)
							left += e.getValue()[0];
						else
							right += e.getValue()[0];
					}
					left = sum - left;
					right = sum - right;
					if (0 > left && left >= -0.0001)
						left = 0;
					if (0 > right && right >= -0.0001)
						right = 0;
					assert left >= 0 && right >= 0;

					int colNum = 0;
					for (Map.Entry<Context, Map<Expression, Collection<Element>>> e : ect.cols
							.entrySet()) {
						int colsum = 0;
						for (Collection<Element> cell : e.getValue().values()) {
							int a = cell.size();
							colsum += a;
							lhg_2 -= xlogx(a);
						}
						assert colsum != 0;
						Context cont = e.getKey();
						double colsum0;
						if (cont.sym == null) {
							if (cont.contextType == Context.LEFT_NEIGHBOR)
								colsum0 = left;
							else
								colsum0 = right;
							colNum++;
						} else {
							double[] tmp = ect0.get(e.getKey());
							colsum0 = tmp == null ? 0 : tmp[0];
							if (tmp == null)
								colNum++;
						}
						lhg_2 -= xlogx(colsum0);
						lhg_2 += xlogx(colsum0 + colsum);
					}

					double normalizing2 = NORMALIZATION_MODE == 1 ? rsum * 2
							: ect.rows.size() * (colNum + ect0.size());
					if (lhg_2 < -delta2 * normalizing2)
						continue;
					/*
					 * TODO Enhancement: We may check posterior gain instead of
					 * the above.
					 */

					if (DEBUG_MODE)
						debug_mg_nPastCheck2++;

					/* Merge */
					// 1. merge the AND node
					OrNode mergeTarget = ii == 0 ? and.left : and.right;
					mergeTarget.children.add(n);
					n.parents.add(mergeTarget);
					for (Map.Entry<Symbol, int[]> e : newRowStat.entrySet()) {
						int index = or.children.indexOf(e.getKey());
						or.weights.get(index)[0] += e.getValue()[0];
					}
					double[] tmp = { rsum };
					mergeTarget.weights.add(tmp);
					and.appearanceNum += rsum;

					// 1.1. update the total posterior
					totalPosteriorGain += lhg_1 + lhg_2 / 2;
					// TODO v3 del 2
					/*
					 * lhg_2 here is not the actual value of the 2nd part of
					 * lhg. But in v3 it is.
					 */
					totalPosteriorGain += alpha * (4 * rsum - 2);

					// 2. extract the to-be-reduced part from rowOrCol
					for (Map.Entry<Expression, Map<Context, Collection<Element>>> e : ect.rows
							.entrySet()) {
						Symbol s = ii == 0 ? e.getKey().s2 : e.getKey().s1;
						SymbolSymbolCell apps0 = rowOrCol.get(s);
						apps0.removeAll(e.getValue());
						if (apps0.isEmpty()) {
							if (ii == 0)
								ssm.delCell(n, s);
							else
								ssm.delCell(s, n);
						}
					}

					// 3. do reduction
					doReduction(ect, and);

					assert ssm.isConsistent();

					// 4. do maximal parsing
					for (Map.Entry<Context, Map<Expression, Collection<Element>>> e : ect.cols
							.entrySet()) {
						if (e.getKey().contextType == Context.LEFT_NEIGHBOR)
							for (Collection<Element> apps : e.getValue()
									.values()) {
								for (Element element : apps) {
									/*
									 * It's possible that the element has
									 * already been reduced when max-reducing
									 * some other element in this for-loop. If
									 * so, either the element's symbol field is
									 * no longer "and", or the element is no
									 * longer in the sentence (i.e., its
									 * prev.next is not itself).
									 */
									/*
									 * There's actually a special case besides
									 * the above two possibilities, that some
									 * previous max-reduction has reduced this
									 * element, but this element happens to be
									 * the left-most element being reduced and
									 * the max-reduction root happens to be
									 * "and". There's no simple way to detect
									 * this special case, but fortunately, the
									 * code below still holds in the special
									 * case and the only downside is that it
									 * wastes some time.
									 */
									if (element.sym != and)
										continue;
									if (element.prev == null) {
										if (element.next == null)
											continue; // parsing done
										else
											assert element.next.prev == element;
									} else if (element.prev.next != element)
										continue;

									Element newElem = doMaxReduction(element);
									if (newElem.prev != null) {
										SymbolSymbolCell cell = ssm.getCell(
												newElem.prev.sym, newElem.sym);
										if (cell != null)
											cell.add(newElem.prev);
										else {
											cell = new SymbolSymbolCell();
											cell.add(newElem.prev);
											if (cf.isConstituent(cell)) {
												ssm.addCell(newElem.prev.sym,
														newElem.sym, cell);
											}
										}
									}
									if (newElem.next != null) {
										SymbolSymbolCell cell = ssm.getCell(
												newElem.sym, newElem.next.sym);
										if (cell != null)
											cell.add(newElem);
										else {
											cell = new SymbolSymbolCell();
											cell.add(newElem);
											if (cf.isConstituent(cell)) {
												ssm.addCell(newElem.sym,
														newElem.next.sym, cell);
											}
										}
									}
									assert ssm.isConsistent();
								}
							}
					}
					assert ssm.isConsistent();

					if (DEBUG_MODE) {
						String ect0str = "size: ?*" + ect0.size() + "\n";
						if (ect0.size() <= 50) {
							Object[] cs = ect0.keySet().toArray();
							for (Object ch : cs)
								ect0str += ch + "\t";
							ect0str += "\n";
							for (Object ch : cs) {
								double[] v = ect0.get(ch);
								if (v != null)
									ect0str += v[0];
								ect0str += "\t";
							}
						}
						debug_learning_history.add("Merge to the "
								+ (ii == 0 ? "left" : "right") + ":\n" + and
								+ "\n\nect0:\n" + ect0str + "\nect:\n" + ect);
						debug_learning_history.add(ssm.toString());
					}
				}
			}
		}

		// Post-processing
		// Construct the start symbol
		HashMap<Symbol, double[]> sSet = new HashMap<Symbol, double[]>();
		int count = 0;
		for (Sentence s : c.sentences) {
			if (s.firstElem.next == null) {
				count++;
				double[] v = sSet.get(s.firstElem.sym);
				if (v == null) {
					double[] tmp = { 1 };
					sSet.put(s.firstElem.sym, tmp);
				} else
					v[0]++;
			}
		}
		if (DEBUG_MODE)
			System.out.println("Corpus full-reduction ratio: "
					+ ((double) count) / c.sentences.length);

		ArrayList<Symbol> children = new ArrayList<Symbol>(sSet.size());
		ArrayList<double[]> weights = new ArrayList<double[]>(sSet.size());
		for (Map.Entry<Symbol, double[]> e : sSet.entrySet()) {
			children.add(e.getKey());
			weights.add(e.getValue());
		}
		g.startSymbol = new OrNode(children, weights);
		g.orNodes.add(g.startSymbol);

		/*
		 * TODO Enhancement: Merge similar OR nodes. This can further generalize
		 * the grammar. Should we even do this within the loop? Notice that this
		 * allows an OrNode to have multiple AndNode parents. The code in this
		 * class assumes that an OrNode has exactly one parent, and thus must be
		 * modified!
		 */
		/*
		 * TODO Enhancement: Refine the grammar by e.g., EM. Also notice that if
		 * not all the sentences are reduced to the start-symbol in the end of
		 * the learning, then: 1) there might be sentences in the corpus that
		 * are not reducible to the start-symbol, which we may want to make
		 * reducible; 2) there might be AND nodes that are not reachable from
		 * the start-symbol, which we should either discard or extend.
		 */

		return g;
	}

	/**
	 * After reduction, for some symbol pairs the context is changed. This
	 * method updates all the SymbolSymbolCell in ssm whose contexts may be
	 * affected by reducing a pair of symbols to {@code changed}.
	 * 
	 * @param changed
	 */
	protected void updateSSMCellAfterReducing(Element changed) {
		// TODO v3 rewrite
		Element s1, s2;

		for (int i = 0; i < 2; i++) {
			s1 = i == 0 ? changed.prev : changed.next;
			if (s1 != null) {
				s2 = i == 0 ? s1.prev : s1.next;
				if (s2 != null) {
					if (i == 0) {
						Element tmp = s1;
						s1 = s2;
						s2 = tmp;
					}
					SymbolSymbolCell c = ssm.getCell(s1.sym, s2.sym);
					if (c != null) {
						for (Iterator<Map.Entry<Context, Collection<Element>>> iter = c.ecv
								.entrySet().iterator(); iter.hasNext();) {
							Map.Entry<Context, Collection<Element>> e = (Map.Entry<Context, Collection<Element>>) iter
									.next();
							int contType = i == 0 ? Context.RIGHT_NEIGHBOR
									: Context.LEFT_NEIGHBOR;
							if (e.getKey().contextType == contType) {
								Collection<Element> apps = e.getValue();
								if (apps.remove(s1)) {
									if (apps.isEmpty())
										iter.remove();

									Context cont = new Context(changed.sym,
											contType);
									apps = c.ecv.get(cont);
									if (apps == null) {
										apps = new HashSet<Element>();
										c.ecv.put(cont, apps);
									}
									apps.add(s1);
									break;
								}
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Infer all possible left or right contexts of the AndNode and their
	 * weights (appearance time), and save them in ect0. The contexts should be
	 * symbols in the corpus that is maximally reduced by the current partial
	 * grammar.
	 * <p>
	 * When the weights are saved into ect0, they are multiplied with
	 * {@code coef}, which actually specifies, among all appearances of
	 * {@code and}, the proportion of the AndNode on which this method is
	 * initially called from outside.
	 * 
	 * @param and
	 * @param ect0
	 * @param isLeftContext
	 * @param coef
	 *            should always be 1 if calling from outside this method
	 * @param depth
	 *            the recursive depth, used to prevent cyclic recursion. It
	 *            should be 0 if calling from outside this method
	 */
	protected void inferContext(AndNode and, HashMap<Context, double[]> ect0,
			boolean isLeftContext, double coef, int depth) {
		Map<Symbol, Map<Symbol, SymbolSymbolCell>> rowOrCol = isLeftContext ? ssm.cols
				: ssm.rows;
		Map<Symbol, SymbolSymbolCell> vec = rowOrCol.get(and);
		if (vec != null)
			for (Map.Entry<Symbol, SymbolSymbolCell> e : vec.entrySet()) {
				Context cont = new Context(e.getKey(),
						isLeftContext ? Context.LEFT_NEIGHBOR
								: Context.RIGHT_NEIGHBOR);
				double[] d = ect0.get(cont);
				if (d == null) {
					double[] tmp = { (double) e.getValue().size() * coef };
					ect0.put(cont, tmp);
				} else
					d[0] += e.getValue().size() * coef;
			}

		for (OrNode pa : and.parents) {
			AndNode grandpa = pa.parents.get(0);
			assert grandpa.left != grandpa.right;
			/*
			 * This assertion is a reminder that the following code is based on
			 * the assumption that at this point an AndNode must has two
			 * different OrNode children. When OrNode merging is implemented
			 * within the main loop, this assumption may not be true.
			 */
			boolean isLeftChild = pa == grandpa.left;

			double wsum = 0;
			for (double[] w : pa.weights) {
				wsum += w[0];
			}
			double ratio = pa.weights.get(pa.children.indexOf(and))[0] / wsum;

			if (isLeftChild == isLeftContext) {
				// recursively get the contexts of and's AND parent
				if (depth < icMaxRecursiveDepth)
					inferContext(grandpa, ect0, isLeftContext, coef * ratio,
							depth + 1);
			} else {
				// and's AND brothers are its contexts
				OrNode uncle = isLeftChild ? grandpa.right : grandpa.left;
				for (int i = 0; i < uncle.children.size(); i++) {
					double value = uncle.weights.get(i)[0] * ratio * coef;
					Context cont = new Context(uncle.children.get(i),
							isLeftChild ? Context.RIGHT_NEIGHBOR
									: Context.LEFT_NEIGHBOR);
					double[] v = ect0.get(cont);
					if (v != null)
						v[0] += value;
					else {
						double[] tmp = { value };
						ect0.put(cont, tmp);
					}
				}
			}
		}
	}

	/**
	 * Do maximal reductions on the element, i.e., keep reducing the element or
	 * the reduction results of the element until no further reduction should be
	 * made, using existing grammar rules. This series of reduction can be
	 * represented by a partial parsing tree, where the root is the element
	 * resulted from the last reduction, which is returned, and the leaves
	 * constitute a sub-string of the sentence before the reductions that
	 * includes the parameter element.
	 * 
	 * When doing reduction, some stored pairs would be invalidated.
	 * 
	 * @param element
	 * @return the element resulted from the last reduction
	 */
	protected Element doMaxReduction(Element elem) {
		// Remove elem from ssm (if any) to simplify the invalidation below
		if (elem.next != null)
			ssm.delAppearance(elem);
		if (elem.prev != null)
			ssm.delAppearance(elem.prev);

		while (true) {
			boolean ok = false; // reduce or not
			for (OrNode orpa : elem.sym.parents) {
				assert orpa.parents.get(0).left != orpa.parents.get(0).right;
				/*
				 * This assertion is a reminder that the following code is based
				 * on the assumption that at this point an AndNode must has two
				 * different OrNode children. When OrNode merging is implemented
				 * within the main loop, this assumption may not be true.
				 */
				boolean left = (orpa == orpa.parents.get(0).left);
				Element elem2 = left ? elem.next : elem.prev;
				if (elem2 != null) {
					OrNode orpa2 = left ? orpa.parents.get(0).right
							: orpa.parents.get(0).left;
					if (orpa2.children.contains(elem2.sym)) {
						// check ECT
						Element lcont = left ? elem.prev : elem2.prev;
						if (lcont != null) {
							ok = isInInferredContext(lcont.sym, orpa.parents
									.get(0), true, 0);
							/*
							 * Notice the checking could be incomplete because
							 * e.g., it's possible that lcont.sym is the left-
							 * most symbol of a to-be-reduced substring, while
							 * the context checking may not encounter the same
							 * partially reduced case.
							 */
						} else
							ok = true;

						Element rcont = left ? elem2.next : elem.next;
						if (ok) {
							if (rcont != null) {
								ok = isInInferredContext(rcont.sym,
										orpa.parents.get(0), false, 0);
								/*
								 * Same notes here as above
								 */
							}
						}
						/*
						 * TODO Enhancement: Once again, here we use a strong
						 * condition that only when both contexts are supported
						 * will we do reduction. We may use the weak condition
						 * (i.e., or) instead.
						 */

						// reduce them
						if (ok) {
							// invalidation
							/*
							 * Notice that: 1) at the beginning of this method,
							 * we made sure that "elem" is not in ssm; 2) in the
							 * loop here, each new value of "elem" is a
							 * reduction result and thus not in ssm.
							 */
							if (left) {
								if (rcont != null)
									ssm.delAppearance(left ? elem2 : elem);
							} else {
								if (lcont != null)
									ssm.delAppearance(lcont);
							}

							// add 1 to the corresponding AND/OR weights
							orpa.parents.get(0).appearanceNum++;
							orpa.weights.get(orpa.children.indexOf(elem.sym))[0]++;
							orpa2.weights
									.get(orpa2.children.indexOf(elem2.sym))[0]++;

							// reduction
							Element le = left ? elem : elem2;
							Element re = le.next;
							le.sym = orpa.parents.get(0);
							le.next = re.next;
							if (re.next != null)
								re.next.prev = le;

							// go next
							elem = le;
							break;
						}
					}
				}
			}

			if (!ok) {
				updateSSMCellAfterReducing(elem);
				return elem;
			}
		}
	}

	/**
	 * Check whether {@code sym} is in the left/right context of any possible
	 * appearance of {@code and} in the current (partially reduced) corpus.
	 * 
	 * @param sym
	 * @param and
	 * @param isLeftContext
	 * @param depth
	 *            the recursive depth, used to prevent cyclic recursion. It
	 *            should be 0 if calling from outside this method
	 * @return
	 */
	protected boolean isInInferredContext(Symbol sym, AndNode and,
			boolean isLeftContext, int depth) {
		Map<Symbol, Map<Symbol, SymbolSymbolCell>> rowOrCol = isLeftContext ? ssm.cols
				: ssm.rows;
		Map<Symbol, SymbolSymbolCell> vec = rowOrCol.get(and);
		if (vec != null)
			if (vec.containsKey(sym))
				return true;

		for (OrNode pa : and.parents) {
			AndNode grandpa = pa.parents.get(0);
			assert grandpa.left != grandpa.right;
			/*
			 * This assertion is a reminder that the following code is based on
			 * the assumption that at this point an AndNode must has two
			 * different OrNode children. When OrNode merging is implemented
			 * within the main loop, this assumption may not be true.
			 */
			boolean isLeftChild = pa == grandpa.left;

			if (isLeftChild == isLeftContext) {
				// recursively query and's AND parent
				if (depth < icMaxRecursiveDepth)
					if (isInInferredContext(sym, grandpa, isLeftContext,
							depth + 1))
						return true;
			} else {
				// and's AND brothers are its contexts
				OrNode uncle = isLeftChild ? grandpa.right : grandpa.left;
				if (uncle.children.contains(sym))
					return true;
			}
		}

		return false;
	}

	/**
	 * Try to find a bi-cluster from the symbol-symbol matrix, using
	 * {@code startBC} as a seed. Overlapping is considered so that the returned
	 * bi-cluster doesn't contain appearances that are considered to belong to
	 * some overlapping bi-clusters.
	 * 
	 * @param startBC
	 *            The seed from which we do bi-clustering. startBC must be
	 *            valid, i.e., it only contains rows/cols presented in ssm. We
	 *            only use startBC.rows/cols, so it's safe if the other fields
	 *            of startBC are not consistent with ssm. Notice that startBC
	 *            will be modified in this method.
	 * @return a bi-cluster, or {@code null} if failed to find a bi-cluster in
	 *         this trial.
	 */
	protected BiCluster biClustering(BiCluster startBC) {
		if (ssm.isEmpty())
			return null;

		/*
		 * We start from a seed and use stochastic hill-climbing to add/del
		 * row/col until a max posterior bi-cluster is found.
		 */
		Symbol[] rows = ssm.rows.keySet().toArray(new Symbol[0]);
		Symbol[] cols = ssm.cols.keySet().toArray(new Symbol[0]);

		/* Seeding */
		BiCluster bc = startBC;
		bc.sum = 0;
		Map<Context, int[]> ectColSums = new HashMap<Context, int[]>();
		for (Map.Entry<Symbol, int[]> e1 : bc.rows.entrySet()) {
			int rowsum = 0;
			for (Symbol s2 : bc.cols.keySet()) {
				SymbolSymbolCell apps = ssm.getCell(e1.getKey(), s2);
				if (apps != null) {
					rowsum += apps.size();

					// construct ectColSums by the way
					for (Map.Entry<Context, Collection<Element>> e : apps.ecv
							.entrySet()) {
						int[] colsum = ectColSums.get(e.getKey());
						if (colsum == null) {
							int[] tmp = { 0 };
							colsum = tmp;
							ectColSums.put(e.getKey(), colsum);
						}
						colsum[0] += e.getValue().size();
					}
				}
			}
			e1.getValue()[0] = rowsum;
			bc.sum += rowsum;
		}
		for (Map.Entry<Symbol, int[]> e2 : bc.cols.entrySet()) {
			int colsum = 0;
			for (Symbol s1 : bc.rows.keySet()) {
				SymbolSymbolCell apps = ssm.getCell(s1, e2.getKey());
				if (apps != null)
					colsum += apps.size();
			}
			e2.getValue()[0] = colsum;
		}

		/* Stochastic Hill-climbing */
		double[] r_contributions = new double[rows.length];
		double[] c_contributions = new double[cols.length];
		int[] r_sums = new int[rows.length];
		int[] c_sums = new int[cols.length];
		ArrayList<Map<Context, int[]>> r_EctColSums = new ArrayList<Map<Context, int[]>>(
				rows.length);
		ArrayList<Map<Context, int[]>> c_EctColSums = new ArrayList<Map<Context, int[]>>(
				cols.length);

		// used for binary search of the next step
		double[] weightSums = new double[rows.length + cols.length + 1];
		weightSums[0] = 0;
		int[] candidateIndex = new int[rows.length + cols.length + 1];
		int weightSumsInd = 0;
		int rowColBoundaryInd; // the last row candidate

		// pre-compute the contribution of all possible steps
		for (int i = 0; i < rows.length; i++) {
			Symbol s = rows[i];
			boolean inBC = bc.rows.containsKey(s);

			// r_sums
			if (inBC)
				r_sums[i] = bc.rows.get(s)[0];
			else {
				r_sums[i] = 0;
				for (Symbol sym : bc.cols.keySet()) {
					SymbolSymbolCell c = ssm.getCell(s, sym);
					if (c != null)
						r_sums[i] += c.size();
				}
			}

			// r_EctColSums
			Map<Context, int[]> newEctColSums = new HashMap<Context, int[]>();
			r_EctColSums.add(newEctColSums);

			// r_contributions: two coherence
			// if (inBC && bc.rows.size() == 1)
			// continue;
			int sign = inBC ? -1 : 1;
			r_contributions[i] = 0;
			int rsum = r_sums[i];
			if (rsum != 0) {
				r_contributions[i] += sign * xlogx(rsum);
				r_contributions[i] += 2 * xlogx(bc.sum);
				r_contributions[i] -= 2 * xlogx(sign * rsum + bc.sum);
				for (Map.Entry<Symbol, int[]> entry : bc.cols.entrySet()) {
					SymbolSymbolCell c = ssm.getCell(s, entry.getKey());
					if (c != null) {
						int a = c.size();
						int l = entry.getValue()[0];
						// r_contributions[i] -= sign * xlogx(a);
						r_contributions[i] -= xlogx(l);
						r_contributions[i] += xlogx(sign * a + l);

						for (Map.Entry<Context, Collection<Element>> e : c.ecv
								.entrySet()) {
							r_contributions[i] -= sign
									* xlogx(e.getValue().size()) / 2;
							// TODO v3 del /2

							int[] colsum = newEctColSums.get(e.getKey());
							if (colsum == null) {
								int[] tmp = { 0 };
								colsum = tmp;
								newEctColSums.put(e.getKey(), colsum);
							}
							colsum[0] += e.getValue().size();
						}
					} // else it's "+=0"
				}
			} // else it's "+=0"
			for (Map.Entry<Context, int[]> e : newEctColSums.entrySet()) {
				int newc = e.getValue()[0];
				int[] c = ectColSums.get(e.getKey());
				if (c != null) {
					r_contributions[i] += xlogx(c[0] + sign * newc) / 2;
					// TODO v3 del /2;
					r_contributions[i] -= xlogx(c[0]) / 2;
					// TODO v3 del /2;
				} else
					r_contributions[i] += xlogx(newc) / 2;
				// TODO v3 del /2;
			}

			// r_contributions: prior change
			r_contributions[i] += sign * 4 * alpha * (rsum - 0.5);

			// weight
			if (inBC && bc.rows.size() == 1)
				continue;
			double weight = shcContrib2Weight(r_contributions[i] / bc.sum, T1);
			if (weight != 0) {
				weightSums[weightSumsInd + 1] = weightSums[weightSumsInd]
						+ weight;
				weightSumsInd++;
				candidateIndex[weightSumsInd] = i;
			}
		}
		rowColBoundaryInd = weightSumsInd;
		for (int i = 0; i < cols.length; i++) {
			Symbol s = cols[i];
			boolean inBC = bc.cols.containsKey(s);

			// c_sums
			if (inBC)
				c_sums[i] = bc.cols.get(s)[0];
			else {
				c_sums[i] = 0;
				for (Symbol sym : bc.rows.keySet()) {
					SymbolSymbolCell c = ssm.getCell(sym, s);
					if (c != null)
						c_sums[i] += c.size();
				}
			}

			// c_EctColSums
			Map<Context, int[]> newEctColSums = new HashMap<Context, int[]>();
			c_EctColSums.add(newEctColSums);

			// c_contributions: two coherence
			// if (inBC && bc.cols.size() == 1)
			// continue;
			int sign = inBC ? -1 : 1;
			c_contributions[i] = 0;
			int csum = c_sums[i];
			if (csum != 0) {
				c_contributions[i] += sign * xlogx(csum);
				c_contributions[i] += 2 * xlogx(bc.sum);
				c_contributions[i] -= 2 * xlogx(sign * csum + bc.sum);
				for (Map.Entry<Symbol, int[]> entry : bc.rows.entrySet()) {
					SymbolSymbolCell c = ssm.getCell(entry.getKey(), s);
					if (c != null) {
						int a = c.size();
						int l = entry.getValue()[0];
						// c_contributions[i] -= sign * xlogx(a);
						c_contributions[i] -= xlogx(l);
						c_contributions[i] += xlogx(sign * a + l);

						for (Map.Entry<Context, Collection<Element>> e : c.ecv
								.entrySet()) {
							c_contributions[i] -= sign
									* xlogx(e.getValue().size()) / 2;
							// TODO v3 del /2

							int[] colsum = newEctColSums.get(e.getKey());
							if (colsum == null) {
								int[] tmp = { 0 };
								colsum = tmp;
								newEctColSums.put(e.getKey(), colsum);
							}
							colsum[0] += e.getValue().size();
						}
					} // else it's "+=0"
				}
			} // else it's "+=0"
			for (Map.Entry<Context, int[]> e : newEctColSums.entrySet()) {
				int newc = e.getValue()[0];
				int[] c = ectColSums.get(e.getKey());
				if (c != null) {
					c_contributions[i] += xlogx(c[0] + sign * newc) / 2;
					// TODO v3 del /2
					c_contributions[i] -= xlogx(c[0]) / 2;
					// TODO v3 del /2
				} else
					c_contributions[i] += xlogx(newc) / 2;
				// TODO v3 del /2
			}

			// c_contributions: prior change
			c_contributions[i] += sign * 4 * alpha * (csum - 0.5);

			// weight
			if (inBC && bc.cols.size() == 1)
				continue;
			double weight = shcContrib2Weight(c_contributions[i] / bc.sum, T1);
			if (weight != 0) {
				weightSums[weightSumsInd + 1] = weightSums[weightSumsInd]
						+ weight;
				weightSumsInd++;
				candidateIndex[weightSumsInd] = i;
			}
		}

		while (true) {
			if (weightSumsInd == 0)
				break; // local max reached
			else {
				// choose the next step
				double rand = Math.random() * weightSums[weightSumsInd];
				int ind = Arrays.binarySearch(weightSums, 1, weightSumsInd,
						rand);
				if (ind < 0)
					ind = -ind - 1;
				boolean nextStepIsRow = ind <= rowColBoundaryInd;
				int nextStepIndex = candidateIndex[ind];
				Symbol nextStepSymbol = nextStepIsRow ? rows[nextStepIndex]
						: cols[nextStepIndex];
				boolean nextStepIsAdding = nextStepIsRow ? !bc.rows
						.containsKey(nextStepSymbol) : !bc.cols
						.containsKey(nextStepSymbol);

				// reset binary search variables
				weightSumsInd = 0;

				// update the contribution of all possible steps rows
				int bSign = nextStepIsAdding ? 1 : -1;
				for (int i = 0; i < rows.length; i++) {
					// a for rows[i], b for nextStepSymbol
					int aSign = bc.rows.containsKey(rows[i]) ? -1 : 1;

					boolean allowed = true;
					// update contribution
					if (nextStepIsRow) {
						if (rows[i] == nextStepSymbol) {
							r_contributions[i] = -r_contributions[i];
						} else {
							/*
							 * TODO Enhancement: If ssm is sparse, we can make
							 * this block more efficient. Currently, for each
							 * row we check each of the bi-cluster columns.
							 * Instead, we can go over each bi-cluster column
							 * and update the involved rows, which is faster
							 * because we don't have to go over all rows.
							 */

							if (bc.rows.size() == 2 && aSign < 0 && bSign < 0)
								allowed = false;

							// compute contribution change
							int aSum = r_sums[i];
							int bSum = r_sums[nextStepIndex] * bSign;
							if (aSum != 0 && bSum != 0) {
								r_contributions[i] += 2 * xlogx(bc.sum + bSum);
								r_contributions[i] -= 2 * xlogx(bc.sum + bSum
										+ aSign * aSum);
								r_contributions[i] -= 2 * xlogx(bc.sum);
								r_contributions[i] += 2 * xlogx(bc.sum + aSign
										* aSum);

								for (Map.Entry<Symbol, int[]> e : bc.cols
										.entrySet()) {
									SymbolSymbolCell c = ssm.getCell(rows[i], e
											.getKey());
									if (c != null) {
										int a = c.size();
										c = ssm.getCell(nextStepSymbol, e
												.getKey());
										if (c != null) {
											int b = c.size() * bSign;
											int l = e.getValue()[0];
											r_contributions[i] -= xlogx(l + b);
											r_contributions[i] += xlogx(aSign
													* a + b + l);
											r_contributions[i] += xlogx(l);
											r_contributions[i] -= xlogx(aSign
													* a + l);
										} // else it's +=0
									} // else it's +=0
								}

								Map<Context, int[]> nextStepEctColSums = r_EctColSums
										.get(nextStepIndex);
								for (Map.Entry<Context, int[]> e : r_EctColSums
										.get(i).entrySet()) {
									int[] tmp = nextStepEctColSums.get(e
											.getKey());
									if (tmp != null) {
										int cb = tmp[0] * bSign;
										tmp = ectColSums.get(e.getKey());
										int c = tmp == null ? 0 : tmp[0];
										r_contributions[i] -= xlogx(c + cb) / 2;
										r_contributions[i] += xlogx(aSign
												* e.getValue()[0] + c + cb) / 2;
										r_contributions[i] += xlogx(c) / 2;
										r_contributions[i] -= xlogx(aSign
												* e.getValue()[0] + c) / 2;
										// TODO v3 del /2
									} // else it's +=0
								}
							} // else it's not changed
						}
					} else {
						if (bc.rows.size() == 1
								&& rows[i] == bc.rows.keySet().iterator()
										.next())
							allowed = false;

						// compute contribution change
						int aSum = r_sums[i];
						int bSum = c_sums[nextStepIndex];
						SymbolSymbolCell cell = ssm.getCell(rows[i],
								nextStepSymbol);
						int x = cell != null ? cell.size() : 0;

						double dContrib = 0;

						// change 1
						/*
						 * the following if's are unnecessary but may accelerate
						 * the computation
						 */
						if (x != 0) {
							dContrib -= aSign * xlogx(aSum);
							dContrib += aSign * xlogx(aSum + bSign * x);
							dContrib -= bSign * xlogx(bSum);
							dContrib += bSign * xlogx(bSum + aSign * x);
						} // else it's "+=0"
						if (aSum != 0) {
							dContrib += 2 * xlogx(bc.sum + aSign * aSum);
							dContrib -= 2 * xlogx(bc.sum);
						} // else it's "+=0"
						if (aSum + bSign * x != 0) {
							dContrib += 2 * xlogx(bc.sum + bSign * bSum);
							dContrib -= 2 * xlogx(bc.sum + aSign * aSum + bSign
									* (bSum + aSign * x));
						} // else it's "+=0"

						// update r_sum
						r_sums[i] += bSign * x;

						// change 2
						Map<Context, int[]> aEctColSums = r_EctColSums.get(i);
						for (Map.Entry<Context, int[]> e : aEctColSums
								.entrySet()) {
							int[] tmp = ectColSums.get(e.getKey());
							int c = tmp == null ? 0 : tmp[0];
							dContrib -= xlogx(c + aSign * e.getValue()[0]) / 2;
							dContrib += xlogx(c) / 2;
							// TODO v3 del /2
						}

						// update r_EctColSums
						if (cell != null)
							for (Map.Entry<Context, Collection<Element>> e : cell.ecv
									.entrySet()) {
								int[] colsum = aEctColSums.get(e.getKey());
								if (colsum == null) {
									int[] tmp = { 0 };
									colsum = tmp;
									aEctColSums.put(e.getKey(), colsum);
								}
								colsum[0] += bSign * e.getValue().size();
								assert colsum[0] >= 0;
								if (colsum[0] == 0)
									aEctColSums.remove(e.getKey());

								// btw, change 2 cont'd
								dContrib -= aSign * bSign
										* xlogx(e.getValue().size()) / 2;
								// TODO v3 del /2
							}

						// change 2 cont'd
						Map<Context, int[]> bEctColSums = c_EctColSums
								.get(nextStepIndex);
						for (Map.Entry<Context, int[]> e : aEctColSums
								.entrySet()) {
							int[] tmp = ectColSums.get(e.getKey());
							int c = tmp == null ? 0 : tmp[0];
							tmp = bEctColSums.get(e.getKey());
							int cb = tmp == null ? 0 : tmp[0] * bSign;
							dContrib += xlogx(c + cb + aSign * e.getValue()[0]) / 2;
							dContrib -= xlogx(c + cb) / 2;
							// TODO v3 del /2
						}

						// change 3: prior change
						dContrib += aSign * 4 * alpha * bSign * x;

						// update contribution
						r_contributions[i] += dContrib;

						// } else {
						// r_sums[i] += bSign * c_sums[nextStepIndex];
						// Map<Context, int[]> aEctColSums = r_EctColSums
						// .get(i);
						// for (Map.Entry<Context, int[]> e : c_EctColSums
						// .get(nextStepIndex).entrySet()) {
						// int[] colsum = aEctColSums.get(e.getKey());
						// if (colsum == null) {
						// int[] tmp = { 0 };
						// colsum = tmp;
						// aEctColSums.put(e.getKey(), colsum);
						// }
						// colsum[0] += bSign * e.getValue()[0];
						// assert colsum[0] >= 0;
						// if (colsum[0] == 0)
						// aEctColSums.remove(e.getKey());
						// }
						// continue;
						// }
					}

					// set binary search variables
					if (!allowed)
						continue;
					double weight = shcContrib2Weight(r_contributions[i]
							/ bc.sum, T1);
					if (weight != 0) {
						weightSums[weightSumsInd + 1] = weightSums[weightSumsInd]
								+ weight;
						weightSumsInd++;
						candidateIndex[weightSumsInd] = i;
					}
				}
				// cols
				rowColBoundaryInd = weightSumsInd;
				for (int i = 0; i < cols.length; i++) {
					// a for cols[i], b for nextStepSymbol
					int aSign = bc.cols.containsKey(cols[i]) ? -1 : 1;

					boolean allowed = true;
					// update contribution
					if (!nextStepIsRow) {
						if (cols[i] == nextStepSymbol) {
							c_contributions[i] = -c_contributions[i];
						} else {
							/*
							 * TODO Enhancement: See the to-do enhancement
							 * above.
							 */

							if (bc.cols.size() == 2 && aSign < 0 && bSign < 0)
								allowed = false;

							// compute contribution change
							int aSum = c_sums[i];
							int bSum = c_sums[nextStepIndex] * bSign;
							if (aSum != 0 && bSum != 0) {
								c_contributions[i] += 2 * xlogx(bc.sum + bSum);
								c_contributions[i] -= 2 * xlogx(bc.sum + bSum
										+ aSign * aSum);
								c_contributions[i] -= 2 * xlogx(bc.sum);
								c_contributions[i] += 2 * xlogx(bc.sum + aSign
										* aSum);

								for (Map.Entry<Symbol, int[]> e : bc.rows
										.entrySet()) {
									SymbolSymbolCell c = ssm.getCell(
											e.getKey(), cols[i]);
									if (c != null) {
										int a = c.size();
										c = ssm.getCell(e.getKey(),
												nextStepSymbol);
										if (c != null) {
											int b = c.size() * bSign;
											int l = e.getValue()[0];
											c_contributions[i] -= xlogx(l + b);
											c_contributions[i] += xlogx(aSign
													* a + b + l);
											c_contributions[i] += xlogx(l);
											c_contributions[i] -= xlogx(aSign
													* a + l);
										} // else it's +=0
									} // else it's +=0
								}

								Map<Context, int[]> nextStepEctColSums = c_EctColSums
										.get(nextStepIndex);
								for (Map.Entry<Context, int[]> e : c_EctColSums
										.get(i).entrySet()) {
									int[] tmp = nextStepEctColSums.get(e
											.getKey());
									if (tmp != null) {
										int cb = tmp[0] * bSign;
										tmp = ectColSums.get(e.getKey());
										int c = tmp == null ? 0 : tmp[0];
										c_contributions[i] -= xlogx(c + cb) / 2;
										c_contributions[i] += xlogx(aSign
												* e.getValue()[0] + c + cb) / 2;
										c_contributions[i] += xlogx(c) / 2;
										c_contributions[i] -= xlogx(aSign
												* e.getValue()[0] + c) / 2;
										// TODO v3 del /2
									} // else it's +=0
								}
							} // else it's not changed
						}
					} else {
						if (bc.cols.size() == 1
								&& cols[i] == bc.cols.keySet().iterator()
										.next())
							allowed = false;

						// compute contribution change
						int aSum = c_sums[i];
						int bSum = r_sums[nextStepIndex];
						SymbolSymbolCell cell = ssm.getCell(nextStepSymbol,
								cols[i]);
						int x = cell != null ? cell.size() : 0;

						double dContrib = 0;

						// change 1
						/*
						 * the following if's are unnecessary but may accelerate
						 * the computation
						 */
						if (x != 0) {
							dContrib -= aSign * xlogx(aSum);
							dContrib += aSign * xlogx(aSum + bSign * x);
							dContrib -= bSign * xlogx(bSum);
							dContrib += bSign * xlogx(bSum + aSign * x);
						} // else it's "+=0"
						if (aSum != 0) {
							dContrib += 2 * xlogx(bc.sum + aSign * aSum);
							dContrib -= 2 * xlogx(bc.sum);
						} // else it's "+=0"
						if (aSum + bSign * x != 0) {
							dContrib += 2 * xlogx(bc.sum + bSign * bSum);
							dContrib -= 2 * xlogx(bc.sum + aSign * aSum + bSign
									* (bSum + aSign * x));
						} // else it's "+=0"

						// update c_sum
						c_sums[i] += bSign * x;

						// change 2
						Map<Context, int[]> aEctColSums = c_EctColSums.get(i);
						for (Map.Entry<Context, int[]> e : aEctColSums
								.entrySet()) {
							int[] tmp = ectColSums.get(e.getKey());
							int c = tmp == null ? 0 : tmp[0];
							dContrib -= xlogx(c + aSign * e.getValue()[0]) / 2;
							dContrib += xlogx(c) / 2;
							// TODO v3 del /2
						}

						// update c_EctColSums
						if (cell != null)
							for (Map.Entry<Context, Collection<Element>> e : cell.ecv
									.entrySet()) {
								int[] colsum = aEctColSums.get(e.getKey());
								if (colsum == null) {
									int[] tmp = { 0 };
									colsum = tmp;
									aEctColSums.put(e.getKey(), colsum);
								}
								colsum[0] += bSign * e.getValue().size();
								assert colsum[0] >= 0;
								if (colsum[0] == 0)
									aEctColSums.remove(e.getKey());

								// btw, change 2 cont'd
								dContrib -= aSign * bSign
										* xlogx(e.getValue().size()) / 2;
								// TODO v3 del /2
							}

						// change 2 cont'd
						Map<Context, int[]> bEctColSums = r_EctColSums
								.get(nextStepIndex);
						for (Map.Entry<Context, int[]> e : aEctColSums
								.entrySet()) {
							int[] tmp = ectColSums.get(e.getKey());
							int c = tmp == null ? 0 : tmp[0];
							tmp = bEctColSums.get(e.getKey());
							int cb = tmp == null ? 0 : tmp[0] * bSign;
							dContrib += xlogx(c + cb + aSign * e.getValue()[0]) / 2;
							dContrib -= xlogx(c + cb) / 2;
							// TODO v3 del /2
						}

						// change 3: prior change
						dContrib += aSign * 4 * alpha * bSign * x;

						// update contribution
						c_contributions[i] += dContrib;

						// } else {
						// c_sums[i] += bSign * r_sums[nextStepIndex];
						// Map<Context, int[]> aEctColSums = c_EctColSums
						// .get(i);
						// for (Map.Entry<Context, int[]> e : r_EctColSums
						// .get(nextStepIndex).entrySet()) {
						// int[] colsum = aEctColSums.get(e.getKey());
						// if (colsum == null) {
						// int[] tmp = { 0 };
						// colsum = tmp;
						// aEctColSums.put(e.getKey(), colsum);
						// }
						// colsum[0] += bSign * e.getValue()[0];
						// assert colsum[0] >= 0;
						// if (colsum[0] == 0)
						// aEctColSums.remove(e.getKey());
						// }
						// continue;
						// }
					}

					// set binary search variables
					if (!allowed)
						continue;
					double weight = shcContrib2Weight(c_contributions[i]
							/ bc.sum, T1);
					if (weight != 0) {
						weightSums[weightSumsInd + 1] = weightSums[weightSumsInd]
								+ weight;
						weightSumsInd++;
						candidateIndex[weightSumsInd] = i;
					}
				}

				// update bc
				if (nextStepIsRow) {
					if (nextStepIsAdding) {
						int[] tmp = { r_sums[nextStepIndex] };
						bc.rows.put(nextStepSymbol, tmp);
					} else
						bc.rows.remove(nextStepSymbol);
					bc.sum += bSign * r_sums[nextStepIndex];
					for (Map.Entry<Symbol, int[]> entry : bc.cols.entrySet()) {
						SymbolSymbolCell c = ssm.getCell(nextStepSymbol, entry
								.getKey());
						int b = (c != null ? c.size() : 0) * bSign;
						entry.getValue()[0] += b;
					}
					for (Map.Entry<Context, int[]> e : r_EctColSums.get(
							nextStepIndex).entrySet()) {
						int[] colsum = ectColSums.get(e.getKey());
						if (colsum == null) {
							int[] tmp = { 0 };
							colsum = tmp;
							ectColSums.put(e.getKey(), colsum);
						}
						colsum[0] += bSign * e.getValue()[0];
						assert colsum[0] >= 0;
						if (colsum[0] == 0)
							ectColSums.remove(e.getKey());
					}
				} else {
					if (nextStepIsAdding) {
						int[] tmp = { c_sums[nextStepIndex] };
						bc.cols.put(nextStepSymbol, tmp);
					} else
						bc.cols.remove(nextStepSymbol);
					bc.sum += bSign * c_sums[nextStepIndex];
					for (Map.Entry<Symbol, int[]> entry : bc.rows.entrySet()) {
						SymbolSymbolCell c = ssm.getCell(entry.getKey(),
								nextStepSymbol);
						int b = (c != null ? c.size() : 0) * bSign;
						entry.getValue()[0] += b;
					}
					for (Map.Entry<Context, int[]> e : c_EctColSums.get(
							nextStepIndex).entrySet()) {
						int[] colsum = ectColSums.get(e.getKey());
						if (colsum == null) {
							int[] tmp = { 0 };
							colsum = tmp;
							ectColSums.put(e.getKey(), colsum);
						}
						colsum[0] += bSign * e.getValue()[0];
						assert colsum[0] >= 0;
						if (colsum[0] == 0)
							ectColSums.remove(e.getKey());
					}
				}
			}
		}
		/*
		 * Note: It is guaranteed that the resulting bi-cluster doesn't contain
		 * any empty row/col.
		 */

		if (DEBUG_MODE)
			debug_bc_nHC1++;

		/* Construct ECT */
		bc.ect = new ExpressionContextMatrix();
		for (Symbol s1 : bc.rows.keySet()) {
			for (Symbol s2 : bc.cols.keySet()) {
				SymbolSymbolCell c = ssm.getCell(s1, s2);
				if (c != null) {
					bc.ect.addRow(s1, s2, c);
				}
			}
		}

		if (DEBUG_MODE) {
			debug_learning_history.add(bc.toString());
			debug_learning_history.add(bc.ect.toString());
		}

		/* Overlapping filtering */
		/*
		 * TODO Enhancement: The overlapping checking here only looks at those
		 * cells whose appearance# are much higher than the prediction by the
		 * multiplicative coherence. In this way, a one-row or one-col
		 * bi-cluster won't be checked at all. One way that can also check
		 * one-row/col bi-clusters is to look at each non-zero cell in ECT, that
		 * if it's the only non-zero cell in the column, then it may be filtered
		 * (when the corpus is large enough). However this second way can't find
		 * the kind of overlapping where multiple cells belong to the same other
		 * bi-cluster. So these two way should be somehow combined, e.g., in
		 * sequence.
		 */
		if (DEBUG_MODE)
			debug_of_bc_nBeforeOF += bc.sum;
		HashSet<Expression> abnormals = new HashSet<Expression>();
		// cells that need to be checked for overlapping
		double mult = mcbcMaxRatioToPrediction / bc.sum;
		for (Map.Entry<Symbol, int[]> r : bc.rows.entrySet()) {
			for (Map.Entry<Symbol, int[]> c : bc.cols.entrySet()) {
				SymbolSymbolCell apps = ssm.getCell(r.getKey(), c.getKey());
				if (apps != null)
					if (apps.size() > mult * r.getValue()[0] * c.getValue()[0])
						abnormals.add(new Expression(r.getKey(), c.getKey()));
			}
		}
		for (Expression exp : abnormals) {
			Map<Context, Collection<Element>> r = bc.ect.rows.get(exp);
			if (r != null) {
				Context[] conts = r.keySet().toArray(new Context[0]);
				for (int i = 0; i < conts.length; i++) {
					Context cont = conts[i];
					Map<Expression, Collection<Element>> column = bc.ect.cols
							.get(cont);
					if (column != null)
						if (abnormals.containsAll(column.keySet())) {
							for (Map.Entry<Expression, Collection<Element>> entry : column
									.entrySet()) {
								Expression ex = entry.getKey();
								Collection<Element> co = entry.getValue();

								// TODO v3 below is not needed
								for (Element elem : co) {
									boolean left = (cont.contextType == Context.LEFT_NEIGHBOR);
									Element neighbor2 = left ? elem.next.next
											: elem.prev;
									/*
									 * TODO Enhancement: Again, here we use a
									 * strong condition that when either side of
									 * the context meets the condition, the
									 * appearance is filtered. And a less strict
									 * alternative is using an AND condition.
									 * Though, here the ECT is more complete
									 * than the previous two cases, so using the
									 * strong condition is more justified.
									 */
									Symbol s = neighbor2 != null ? neighbor2.sym
											: null;
									Context cont2 = new Context(s,
											left ? Context.RIGHT_NEIGHBOR
													: Context.LEFT_NEIGHBOR);
									bc.ect.delAppearance(ex, cont2, elem);

									int[] tmp = ectColSums.get(cont2);
									tmp[0]--;
									if (tmp[0] == 0)
										ectColSums.remove(cont2);
								}

								// update bc
								bc.sum -= co.size();
								if ((bc.rows.get(ex.s1)[0] -= co.size()) == 0)
									bc.rows.remove(ex.s1);
								if ((bc.cols.get(ex.s2)[0] -= co.size()) == 0)
									bc.cols.remove(ex.s2);
							}
							bc.ect.delCol(cont);

							ectColSums.remove(cont);

							if (DEBUG_MODE)
								debug_learning_history
										.add("Overlapping-Filtered:\n"
												+ bc.toString());
						}
				}
			}
		}
		if (DEBUG_MODE)
			debug_of_bc_nAfterOF += bc.sum;

		// bicluster coherence checking
		double coherence1 = 0, coherence2 = 0;
		coherence1 -= xlogx(bc.sum);
		coherence2 -= xlogx(bc.sum * 2); // TODO v3 del *2
		for (int[] i : bc.cols.values())
			coherence1 += xlogx(i[0]);
		for (int[] colsum : ectColSums.values()) {
			coherence2 += xlogx(colsum[0]);
		}
		for (Map.Entry<Symbol, int[]> e : bc.rows.entrySet()) {
			coherence1 += xlogx(e.getValue()[0]);

			for (Symbol s2 : bc.cols.keySet()) {
				Map<Context, Collection<Element>> m = bc.ect.rows
						.get(new Expression(e.getKey(), s2));
				if (m != null) {
					int a = 0;
					for (Collection<Element> v : m.values()) {
						int a2 = v.size();
						coherence2 -= xlogx(a2);
						a += a2;
					}
					coherence1 -= xlogx(a / 2); // TODO v3 del /2
					coherence2 += xlogx(a);
				}
			}
		}

		int normalizing1 = NORMALIZATION_MODE == 1 ? bc.sum : bc.rows.size()
				* bc.cols.size();
		int normalizing2 = NORMALIZATION_MODE == 1 ? bc.sum * 2 : bc.ect.rows
				.size()
				* bc.ect.cols.size(); // TODO v3 del *2
		if (-coherence1 > delta * normalizing1
				|| -coherence2 > delta2 * normalizing2) {
			if (DEBUG_MODE)
				debug_bc_nFailedCheck1++;

			/* shrink by hill-climbing to make it coherent */
			// initialization
			Map<Symbol, Map<Symbol, int[]>> bcssm = new HashMap<Symbol, Map<Symbol, int[]>>();
			for (Symbol s1 : bc.rows.keySet()) {
				Map<Symbol, int[]> row = new HashMap<Symbol, int[]>();
				bcssm.put(s1, row);
				for (Symbol s2 : bc.cols.keySet()) {
					Map<Context, Collection<Element>> m = bc.ect.rows
							.get(new Expression(s1, s2));
					if (m != null) {
						int a = 0;
						for (Collection<Element> v : m.values())
							a += v.size();
						int[] tmp = { a / 2 }; // TODO v3 del 2
						row.put(s2, tmp);
					}
				}
			}

			// contributions[]: d-coherence1, d-coherence2, d-prior
			Map<Symbol, double[]> r_contributions2 = new HashMap<Symbol, double[]>(
					(int) (bc.rows.size() / 0.75));
			Map<Symbol, double[]> c_contributions2 = new HashMap<Symbol, double[]>(
					(int) (bc.cols.size() / 0.75));
			// normalizing[]: for mode 2: d-ectRow#, d-ectCol#
			Map<Symbol, int[]> r_normalizing = new HashMap<Symbol, int[]>(
					(int) (bc.rows.size() / 0.75));
			Map<Symbol, int[]> c_normalizing = new HashMap<Symbol, int[]>(
					(int) (bc.cols.size() / 0.75));
			// EctColSums
			Map<Symbol, Map<Context, int[]>> r_EctColSums2 = new HashMap<Symbol, Map<Context, int[]>>(
					bc.rows.size() * 2);
			Map<Symbol, Map<Context, int[]>> c_EctColSums2 = new HashMap<Symbol, Map<Context, int[]>>(
					bc.cols.size() * 2);

			double bestContrib = Double.NEGATIVE_INFINITY;
			Symbol bestCand = null;
			boolean bestCandIsRow = true;

			if (bc.rows.size() > 1)
				for (Map.Entry<Symbol, int[]> e : bc.rows.entrySet()) {
					// initialize the set of variables
					Symbol s = e.getKey();
					int rsum = e.getValue()[0];

					double[] contrib = new double[3];
					r_contributions2.put(s, contrib);

					Map<Context, int[]> ecv = new HashMap<Context, int[]>();
					r_EctColSums2.put(s, ecv);

					int[] ectsize = new int[2];
					r_normalizing.put(s, ectsize);

					contrib[0] -= xlogx(rsum);
					for (Map.Entry<Symbol, int[]> e2 : bc.cols.entrySet()) {
						int[] tmp = bcssm.get(s).get(e2.getKey());
						if (tmp != null) {
							int a = tmp[0];
							contrib[0] += xlogx(e2.getValue()[0] - a);
							contrib[0] -= xlogx(e2.getValue()[0]);
							contrib[0] += xlogx(a);
							contrib[1] -= xlogx(a * 2); // TODO v3 del *2
							/*
							 * Notice that contrib[0]+contrib[1]/2 isn't equal
							 * to the likelihood gain, as here the two terms
							 * don't conteract each other. But still we use it
							 * to approximate the real one. In v3, however, they
							 * are equal.
							 */

							Map<Context, Collection<Element>> ectrow = bc.ect.rows
									.get(new Expression(s, e2.getKey()));
							for (Map.Entry<Context, Collection<Element>> e3 : ectrow
									.entrySet()) {
								int aa = e3.getValue().size();
								contrib[1] += xlogx(aa);

								int[] colsum = ecv.get(e3.getKey());
								if (colsum == null) {
									colsum = new int[1];
									ecv.put(e3.getKey(), colsum);
								}
								colsum[0] += aa;
							}

							ectsize[0]++;
						}
					}
					for (Map.Entry<Context, int[]> e2 : ecv.entrySet()) {
						int c = e2.getValue()[0];
						int[] tmp = ectColSums.get(e2.getKey());
						int c0 = tmp == null ? 0 : tmp[0];
						contrib[1] += xlogx(c0 - c);
						contrib[1] -= xlogx(c0);

						if (c == c0)
							ectsize[1]++;
					}
					contrib[0] += xlogx(bc.sum);
					contrib[0] -= xlogx(bc.sum - rsum);
					contrib[1] += xlogx(2 * bc.sum);
					contrib[1] -= xlogx(2 * (bc.sum - rsum));
					// TODO v3 del *2

					contrib[2] = 4 * alpha * (0.5 - rsum);

					// if it's the best...
					double contribution = contrib[0] + contrib[1] / 2
							+ contrib[2];
					// TODO v3 del /2
					if (contribution > bestContrib) {
						bestContrib = contribution;
						bestCand = s;
						bestCandIsRow = true;
					}
				}
			if (bc.cols.size() > 1)
				for (Map.Entry<Symbol, int[]> e : bc.cols.entrySet()) {
					// initialize the set of variables
					Symbol s = e.getKey();
					int csum = e.getValue()[0];

					double[] contrib = new double[3];
					c_contributions2.put(s, contrib);

					Map<Context, int[]> ecv = new HashMap<Context, int[]>();
					c_EctColSums2.put(s, ecv);

					int[] ectsize = new int[2];
					c_normalizing.put(s, ectsize);

					contrib[0] -= xlogx(csum);
					for (Map.Entry<Symbol, int[]> e2 : bc.rows.entrySet()) {
						int[] tmp = bcssm.get(e2.getKey()).get(s);
						if (tmp != null) {
							int a = tmp[0];
							contrib[0] += xlogx(e2.getValue()[0] - a);
							contrib[0] -= xlogx(e2.getValue()[0]);
							contrib[0] += xlogx(a);
							contrib[1] -= xlogx(a * 2); // TODO v3 del *2
							/*
							 * Notice that contrib[0]+contrib[1]/2 isn't equal
							 * to the likelihood gain, as here the two terms
							 * don't conteract each other. But still we use it
							 * to approximate the real one. In v3, however, they
							 * are equal.
							 */

							Map<Context, Collection<Element>> ectrow = bc.ect.rows
									.get(new Expression(e2.getKey(), s));
							for (Map.Entry<Context, Collection<Element>> e3 : ectrow
									.entrySet()) {
								int aa = e3.getValue().size();
								contrib[1] += xlogx(aa);

								int[] colsum = ecv.get(e3.getKey());
								if (colsum == null) {
									colsum = new int[1];
									ecv.put(e3.getKey(), colsum);
								}
								colsum[0] += aa;
							}

							ectsize[0]++;
						}
					}
					for (Map.Entry<Context, int[]> e2 : ecv.entrySet()) {
						int c = e2.getValue()[0];
						int[] tmp = ectColSums.get(e2.getKey());
						int c0 = tmp == null ? 0 : tmp[0];
						contrib[1] += xlogx(c0 - c);
						contrib[1] -= xlogx(c0);

						if (c == c0)
							ectsize[1]++;
					}
					contrib[0] += xlogx(bc.sum);
					contrib[0] -= xlogx(bc.sum - csum);
					contrib[1] += xlogx(2 * bc.sum);
					contrib[1] -= xlogx(2 * (bc.sum - csum));
					// TODO v3 del *2

					contrib[2] = 4 * alpha * (0.5 - csum);

					// if it's the best...
					double contribution = contrib[0] + contrib[1] / 2
							+ contrib[2];
					// TODO v3 del /2
					if (contribution > bestContrib) {
						bestContrib = contribution;
						bestCand = s;
						bestCandIsRow = false;
					}
				}

			assert bestCand != null;
			boolean done = false;
			while (!done) {
				// check coherence
				double[] contrib0;
				int[] d_ectsize0;
				int dsum0;
				if (bestCandIsRow) {
					contrib0 = r_contributions2.get(bestCand);
					d_ectsize0 = r_normalizing.get(bestCand);
					dsum0 = bc.rows.get(bestCand)[0];
				} else {
					contrib0 = c_contributions2.get(bestCand);
					d_ectsize0 = c_normalizing.get(bestCand);
					dsum0 = bc.cols.get(bestCand)[0];
				}
				coherence1 += contrib0[0];
				coherence2 += contrib0[1];
				normalizing1 = NORMALIZATION_MODE == 1 ? bc.sum - dsum0
						: (bc.rows.size() - (bestCandIsRow ? 1 : 0))
								* (bc.cols.size() - (bestCandIsRow ? 0 : 1));
				normalizing2 = NORMALIZATION_MODE == 1 ? (bc.sum - dsum0) * 2
						: (bc.ect.rows.size() - d_ectsize0[0])
								* (bc.ect.cols.size() - d_ectsize0[1]);
				// TODO v3 del *2

				if (-coherence1 < delta * normalizing1
						&& -coherence2 < delta2 * normalizing2)
					done = true;

				Symbol newBestCand = null;
				boolean newBestCandIsRow = true;
				double newBestContrib = Double.NEGATIVE_INFINITY;
				if (!done) {
					// update contributions
					// rows
					for (Map.Entry<Symbol, int[]> e : bc.rows.entrySet()) {
						Symbol s = e.getKey();
						int dsum = e.getValue()[0];
						double[] contrib = r_contributions2.get(s);
						int[] d_ectsize = r_normalizing.get(s);

						if (bestCandIsRow) {
							if (e.getKey() == bestCand)
								continue;
							if (bc.rows.size() == 2)
								continue;

							contrib[0] += xlogx(bc.sum - dsum0);
							contrib[0] -= xlogx(bc.sum - dsum0 - dsum);
							contrib[0] -= xlogx(bc.sum);
							contrib[0] += xlogx(bc.sum - dsum);
							for (Map.Entry<Symbol, int[]> e2 : bc.cols
									.entrySet()) {
								int[] c = bcssm.get(s).get(e2.getKey());
								if (c != null) {
									int a = c[0];
									c = bcssm.get(bestCand).get(e2.getKey());
									if (c != null) {
										int b = c[0];
										int l = e2.getValue()[0];
										contrib[0] -= xlogx(l - b);
										contrib[0] += xlogx(l - a - b);
										contrib[0] += xlogx(l);
										contrib[0] -= xlogx(l - a);
									} // else it's +=0
								} // else it's +=0
							}

							contrib[1] += xlogx((bc.sum - dsum0) * 2);
							contrib[1] -= xlogx((bc.sum - dsum0 - dsum) * 2);
							contrib[1] -= xlogx(bc.sum * 2);
							contrib[1] += xlogx((bc.sum - dsum) * 2);
							// TODO v3 del *2
							Map<Context, int[]> bestCandEctColSums = r_EctColSums2
									.get(bestCand);
							for (Map.Entry<Context, int[]> e2 : r_EctColSums2
									.get(s).entrySet()) {
								int[] tmp = bestCandEctColSums.get(e2.getKey());
								if (tmp != null) {
									int cb = tmp[0];
									assert cb != 0;
									int ca = e2.getValue()[0];
									int c = ectColSums.get(e2.getKey())[0];
									// it won't be null
									contrib[1] -= xlogx(c - cb);
									contrib[1] += xlogx(c - cb - ca);
									contrib[1] += xlogx(c);
									contrib[1] -= xlogx(c - ca);

									if (c == ca + cb)
										d_ectsize[1]++;
								} // else it's +=0
							}
						} else {
							if (bc.rows.size() == 1)
								continue;

							int[] cell = bcssm.get(s).get(bestCand);
							int x = cell != null ? cell[0] : 0;

							// if this row will become empty, ignore it
							if (x == dsum)
								continue;

							if (x != 0)
								d_ectsize[0]--;
							d_ectsize[1] = 0;

							/*
							 * the following if's are unnecessary but may
							 * accelerate the computation
							 */
							if (x != 0) {
								contrib[0] += xlogx(dsum);
								contrib[0] -= xlogx(dsum - x);
								contrib[0] += xlogx(dsum0);
								contrib[0] -= xlogx(dsum0 - x);
								contrib[0] -= xlogx(x);
								contrib[1] += xlogx(x * 2); // TODO v3 del 2
							} // else it's "+=0"
							if (dsum != 0) {
								contrib[0] += xlogx(bc.sum - dsum);
								contrib[0] -= xlogx(bc.sum);
								contrib[1] += xlogx((bc.sum - dsum) * 2);
								contrib[1] -= xlogx(bc.sum * 2);
								// TODO v3 del *2
							} // else it's "+=0"
							if (dsum - x != 0) {
								contrib[0] += xlogx(bc.sum - dsum0);
								contrib[0] -= xlogx(bc.sum - dsum - dsum0 + x);
								contrib[1] += xlogx((bc.sum - dsum0) * 2);
								contrib[1] -= xlogx((bc.sum - dsum - dsum0 + x) * 2);
								// TODO v3 del *2
							} // else it's "+=0"

							Map<Context, int[]> aEctColSums = r_EctColSums2
									.get(s);
							for (Map.Entry<Context, int[]> e2 : aEctColSums
									.entrySet()) {
								int c = ectColSums.get(e2.getKey())[0];
								// it won't be null
								contrib[1] -= xlogx(c - e2.getValue()[0]);
								contrib[1] += xlogx(c);
							}

							// update r_EctColSums2
							if (x != 0)
								if (cell != null) {
									Map<Context, Collection<Element>> xecv = bc.ect.rows
											.get(new Expression(s, bestCand));
									for (Map.Entry<Context, Collection<Element>> e2 : xecv
											.entrySet()) {
										int[] colsum = aEctColSums.get(e2
												.getKey());
										colsum[0] -= e2.getValue().size();
										assert colsum[0] >= 0;
										if (colsum[0] == 0)
											aEctColSums.remove(e2.getKey());

										// btw...
										contrib[1] -= xlogx(e2.getValue()
												.size());
									}
								}

							Map<Context, int[]> bEctColSums = c_EctColSums2
									.get(bestCand);
							for (Map.Entry<Context, int[]> e2 : aEctColSums
									.entrySet()) {
								int c = ectColSums.get(e2.getKey())[0];
								// it won't be null
								int[] tmp = bEctColSums.get(e2.getKey());
								int cb = tmp == null ? 0 : tmp[0];
								int ca = e2.getValue()[0];
								contrib[1] += xlogx(c - cb - ca);
								contrib[1] -= xlogx(c - cb);

								if (c == ca + cb)
									d_ectsize[1]++;
							}

							contrib[2] += 4 * alpha * x;
						}

						// if it's the best...
						double contribution = contrib[0] + contrib[1] / 2
								+ contrib[2];
						// TODO v3 del /2
						if (contribution > newBestContrib) {
							newBestContrib = contribution;
							newBestCand = s;
							newBestCandIsRow = true;
						}
					}
					// cols
					for (Map.Entry<Symbol, int[]> e : bc.cols.entrySet()) {
						Symbol s = e.getKey();
						int dsum = e.getValue()[0];
						double[] contrib = c_contributions2.get(s);
						int[] d_ectsize = c_normalizing.get(s);

						if (!bestCandIsRow) {
							if (e.getKey() == bestCand)
								continue;
							if (bc.cols.size() == 2)
								continue;

							contrib[0] += xlogx(bc.sum - dsum0);
							contrib[0] -= xlogx(bc.sum - dsum0 - dsum);
							contrib[0] -= xlogx(bc.sum);
							contrib[0] += xlogx(bc.sum - dsum);
							for (Map.Entry<Symbol, int[]> e2 : bc.rows
									.entrySet()) {
								int[] c = bcssm.get(e2.getKey()).get(s);
								if (c != null) {
									int a = c[0];
									c = bcssm.get(e2.getKey()).get(bestCand);
									if (c != null) {
										int b = c[0];
										int l = e2.getValue()[0];
										contrib[0] -= xlogx(l - b);
										contrib[0] += xlogx(l - a - b);
										contrib[0] += xlogx(l);
										contrib[0] -= xlogx(l - a);
									} // else it's +=0
								} // else it's +=0
							}

							contrib[1] += xlogx((bc.sum - dsum0) * 2);
							contrib[1] -= xlogx((bc.sum - dsum0 - dsum) * 2);
							contrib[1] -= xlogx(bc.sum * 2);
							contrib[1] += xlogx((bc.sum - dsum) * 2);
							// TODO v3 del *2
							Map<Context, int[]> bestCandEctColSums = c_EctColSums2
									.get(bestCand);
							for (Map.Entry<Context, int[]> e2 : c_EctColSums2
									.get(s).entrySet()) {
								int[] tmp = bestCandEctColSums.get(e2.getKey());
								if (tmp != null) {
									int cb = tmp[0];
									assert cb != 0;
									int ca = e2.getValue()[0];
									int c = ectColSums.get(e2.getKey())[0];
									// it won't be null
									contrib[1] -= xlogx(c - cb);
									contrib[1] += xlogx(c - cb - ca);
									contrib[1] += xlogx(c);
									contrib[1] -= xlogx(c - ca);

									if (c == ca + cb)
										d_ectsize[1]++;
								} // else it's +=0
							}
						} else {
							if (bc.cols.size() == 1)
								continue;

							int[] cell = bcssm.get(bestCand).get(s);
							int x = cell != null ? cell[0] : 0;

							// if this col will become empty, ignore it
							if (x == dsum)
								continue;

							if (x != 0)
								d_ectsize[0]--;
							d_ectsize[1] = 0;

							/*
							 * the following if's are unnecessary but may
							 * accelerate the computation
							 */
							if (x != 0) {
								contrib[0] += xlogx(dsum);
								contrib[0] -= xlogx(dsum - x);
								contrib[0] += xlogx(dsum0);
								contrib[0] -= xlogx(dsum0 - x);
								contrib[0] -= xlogx(x);
								contrib[1] += xlogx(x * 2); // TODO v3 del 2
							} // else it's "+=0"
							if (dsum != 0) {
								contrib[0] += xlogx(bc.sum - dsum);
								contrib[0] -= xlogx(bc.sum);
								contrib[1] += xlogx((bc.sum - dsum) * 2);
								contrib[1] -= xlogx(bc.sum * 2);
								// TODO v3 del *2
							} // else it's "+=0"
							if (dsum - x != 0) {
								contrib[0] += xlogx(bc.sum - dsum0);
								contrib[0] -= xlogx(bc.sum - dsum - dsum0 + x);
								contrib[1] += xlogx((bc.sum - dsum0) * 2);
								contrib[1] -= xlogx((bc.sum - dsum - dsum0 + x) * 2);
								// TODO v3 del *2
							} // else it's "+=0"

							Map<Context, int[]> aEctColSums = c_EctColSums2
									.get(s);
							for (Map.Entry<Context, int[]> e2 : aEctColSums
									.entrySet()) {
								int c = ectColSums.get(e2.getKey())[0];
								// it won't be null
								contrib[1] -= xlogx(c - e2.getValue()[0]);
								contrib[1] += xlogx(c);
							}

							// update c_EctColSums2
							if (x != 0)
								if (cell != null) {
									Map<Context, Collection<Element>> xecv = bc.ect.rows
											.get(new Expression(bestCand, s));
									for (Map.Entry<Context, Collection<Element>> e2 : xecv
											.entrySet()) {
										int[] colsum = aEctColSums.get(e2
												.getKey());
										colsum[0] -= e2.getValue().size();
										assert colsum[0] >= 0;
										if (colsum[0] == 0)
											aEctColSums.remove(e2.getKey());

										// btw...
										contrib[1] -= xlogx(e2.getValue()
												.size());
									}
								}

							Map<Context, int[]> bEctColSums = r_EctColSums2
									.get(bestCand);
							for (Map.Entry<Context, int[]> e2 : aEctColSums
									.entrySet()) {
								int c = ectColSums.get(e2.getKey())[0];
								// it won't be null
								int[] tmp = bEctColSums.get(e2.getKey());
								int cb = tmp == null ? 0 : tmp[0];
								int ca = e2.getValue()[0];
								contrib[1] += xlogx(c - cb - ca);
								contrib[1] -= xlogx(c - cb);

								if (c == ca + cb)
									d_ectsize[1]++;
							}

							contrib[2] += 4 * alpha * x;
						}

						// if it's the best...
						double contribution = contrib[0] + contrib[1] / 2
								+ contrib[2];
						// TODO v3 del /2
						if (contribution > newBestContrib) {
							newBestContrib = contribution;
							newBestCand = s;
							newBestCandIsRow = false;
						}
					}
				}

				// update bc
				if (bestCandIsRow) {
					bc.rows.remove(bestCand);
					bc.sum -= dsum0;
					for (Iterator<Map.Entry<Symbol, int[]>> iter = bc.cols
							.entrySet().iterator(); iter.hasNext();) {
						Map.Entry<Symbol, int[]> entry = (Map.Entry<Symbol, int[]>) iter
								.next();
						int[] c = bcssm.get(bestCand).get(entry.getKey());
						if (c != null) {
							if ((entry.getValue()[0] -= c[0]) == 0)
								iter.remove();
							bc.ect.delRow(new Expression(bestCand, entry
									.getKey()));
						}
					}
					for (Map.Entry<Context, int[]> e : r_EctColSums2.get(
							bestCand).entrySet()) {
						int[] colsum = ectColSums.get(e.getKey());
						colsum[0] -= e.getValue()[0];
						assert colsum[0] >= 0;
						if (colsum[0] == 0)
							ectColSums.remove(e.getKey());
					}
				} else {
					bc.cols.remove(bestCand);
					bc.sum -= dsum0;
					for (Iterator<Map.Entry<Symbol, int[]>> iter = bc.rows
							.entrySet().iterator(); iter.hasNext();) {
						Map.Entry<Symbol, int[]> entry = (Map.Entry<Symbol, int[]>) iter
								.next();
						int[] c = bcssm.get(entry.getKey()).get(bestCand);
						if (c != null) {
							if ((entry.getValue()[0] -= c[0]) == 0)
								iter.remove();
							bc.ect.delRow(new Expression(entry.getKey(),
									bestCand));
						}
					}
					for (Map.Entry<Context, int[]> e : c_EctColSums2.get(
							bestCand).entrySet()) {
						int[] colsum = ectColSums.get(e.getKey());
						colsum[0] -= e.getValue()[0];
						assert colsum[0] >= 0;
						if (colsum[0] == 0)
							ectColSums.remove(e.getKey());
					}
				}
				if (DEBUG_MODE)
					debug_learning_history.add("hc2 removes "
							+ (bestCandIsRow ? "row " : "col ") + bestCand);

				assert done || newBestCand != null;
				bestCand = newBestCand;
				bestCandIsRow = newBestCandIsRow;
			}
		}

		if (bc.rows.size() == 1 && bc.cols.size() == 1) {
			if (DEBUG_MODE)
				debug_bc_nScBC2++;

			// checking criterion for single cell bi-cluster
			if (bc.sum < oneCellBiClusterThreshold) {
				if (DEBUG_MODE) {
					debug_bc_nFailedCheck3++;
					debug_learning_history.add("Thresholding failed.");
				}
				return null;
			}
		}

		if (DEBUG_MODE) { // consistency checking
			int count = 0;
			int bcsum = 0;
			for (Symbol s1 : bc.rows.keySet()) {
				for (Symbol s2 : bc.cols.keySet()) {
					Map<Context, Collection<Element>> row = bc.ect.rows
							.get(new Expression(s1, s2));
					if (row != null) {
						count++;
						for (Collection<Element> c : row.values()) {
							assert !c.isEmpty();
							bcsum += c.size();
						}
					}
				}
			}
			assert count == bc.ect.rows.size();
			assert bcsum / 2 == bc.sum;
			// TODO v3 del /2
			int tmp = 0;
			for (int[] i : bc.rows.values()) {
				assert i[0] != 0;
				tmp += i[0];
			}
			assert tmp == bc.sum;
			tmp = 0;
			for (int[] i : bc.cols.values()) {
				assert i[0] != 0;
				tmp += i[0];
			}
			assert tmp == bc.sum;
		}

		/* Compute posterior */
		// TODO v3 we can reuse coherence1/2 above to compute posteriorGain
		double likelihoodGain = -2 * xlogx(bc.sum);
		for (int[] i : bc.cols.values())
			likelihoodGain += xlogx(i[0]);
		for (int[] i : bc.rows.values())
			likelihoodGain += xlogx(i[0]);
		for (Map<Expression, Collection<Element>> col : bc.ect.cols.values()) {
			int colsum = 0;
			for (Collection<Element> c : col.values()) {
				int a = c.size();
				colsum += a;
				likelihoodGain -= xlogx(a) / 2;
				// TODO v3 del /2
			}
			likelihoodGain += xlogx(colsum) / 2;
			// TODO v3 del /2
		}
		double priorGain = 2 * alpha
				* (bc.sum * 2 - 4 - bc.rows.size() - bc.cols.size());
		bc.posteriorGain = likelihoodGain + priorGain;
		/*
		 * We may also check posterior gain
		 * 
		 * TODO Enhancement: after alpha changes with corpus size, try the
		 * following
		 */
		// if(bc.posteriorGain < delta)
		// return null;
		if (DEBUG_MODE) {
			debug_learning_history.add("Likelihood gain: " + likelihoodGain
					+ "\nPosterior gain: " + bc.posteriorGain);
		}

		return bc;
	}

	protected ArrayList<BiCluster> biClusterCandidates = new ArrayList<BiCluster>();

	/**
	 * Do multiple bi-clustering on ssm and return the bi-cluster with the
	 * highest posterior gain.
	 * 
	 * @return the best bi-cluster found and extracted from the matrix, or
	 *         {@code null} if no bi-cluster is found.
	 */
	protected BiCluster bestBiClustering() {
		if (ssm.isEmpty())
			return null;

		double bestGain = Double.NEGATIVE_INFINITY;
		BiCluster bestBC = null;
		ArrayList<BiCluster> newCandidates = new ArrayList<BiCluster>(
				numSimultaneousBiClustering);

		// Check existing candidates
		for (BiCluster bc : biClusterCandidates) {
			bc.rows.keySet().retainAll(ssm.rows.keySet());
			if (!bc.rows.isEmpty()) {
				bc.cols.keySet().retainAll(ssm.cols.keySet());
				if (!bc.cols.isEmpty()) {
					bc = biClustering(bc);
					if (bc != null) {
						// check duplication
						boolean dup = false;
						for (BiCluster bc0 : newCandidates) {
							if (bc.equals(bc0)) {
								dup = true;
								break;
							}
						}
						if (!dup) {
							newCandidates.add(bc);
							if (bestGain < bc.posteriorGain) {
								bestGain = bc.posteriorGain;
								bestBC = bc;
							}
						}
					}
				}
			}
		}

		// Find new ones
		Expression[] seeds = ssm.cells.toArray(new Expression[0]);
		int randMax = seeds.length;
		int num = numSimultaneousBiClustering - newCandidates.size();
		for (int i = 0; i < num; i++) {
			boolean fail = true;
			for (int n = 0; n < biClusteringAttemptNum; n++) {
				// seeding
				/*
				 * Notice that, here we only consider seeds outside existing
				 * bi-clusters. This may potentially prevent certain new
				 * bi-clusters from being discovered, although the chance is
				 * small.
				 */
				Expression seed = null;
				boolean outsider = false;
				while (!outsider && randMax > 0) {
					int ind = (int) (Math.random() * randMax);
					seed = seeds[ind];
					seeds[ind] = seeds[randMax - 1];
					randMax--;

					outsider = true;
					for (BiCluster bc : newCandidates) {
						if (bc.rows.containsKey(seed.s1)
								&& bc.cols.containsKey(seed.s2)) {
							outsider = false;
							break;
						}
					}
				}
				if (!outsider && randMax == 0) // no more outsider
					break;

				// biClustering
				BiCluster bc = new BiCluster();
				bc.sum = ssm.getCell(seed.s1, seed.s2).size();
				int[] tmp1 = { bc.sum };
				bc.rows.put(seed.s1, tmp1);
				int[] tmp2 = { bc.sum };
				bc.cols.put(seed.s2, tmp2);
				bc = biClustering(bc);
				if (bc == null)
					continue;

				// check duplication
				boolean dup = false;
				for (BiCluster bc0 : newCandidates) {
					if (bc.equals(bc0)) {
						dup = true;
						break;
					}
				}
				if (!dup) {
					fail = false;
					newCandidates.add(bc);
					if (bestGain < bc.posteriorGain) {
						bestGain = bc.posteriorGain;
						bestBC = bc;
					}
					break;
				} else
					continue;

				/*
				 * Here once a seed is failed, we simply discard it instead of
				 * trying again. This may also prevent new bi-clusters from
				 * being found, although not very likely.
				 */
			}
			if (fail) // can't find bi-cluster
				break;
		}

		// Return the best one
		if (bestBC != null) {
			newCandidates.remove(bestBC);

			// extract bestBC from ssm
			for (Symbol s1 : bestBC.rows.keySet()) {
				for (Symbol s2 : bestBC.cols.keySet()) {
					SymbolSymbolCell apps = ssm.getCell(s1, s2);
					if (apps != null) {
						Map<Context, Collection<Element>> apps2 = bestBC.ect.rows
								.get(new Expression(s1, s2));
						if (apps2 != null) {
							apps.removeAll(apps2);
							if (apps.isEmpty())
								ssm.delCell(s1, s2);
						}
					}
				}
			}

			totalPosteriorGain += bestGain;
		}
		biClusterCandidates = newCandidates;
		return bestBC;
	}

	/**
	 * Used in stochastic hill-climbing to convert the steepness of a uphill
	 * step to the unnormalized probability weight of choosing that step.
	 * 
	 * @param contrib
	 * @param temp
	 *            "temperature"
	 * @return weight >=0
	 */
	protected static double shcContrib2Weight(double contrib, double temp) {
		return contrib > 0 ? Math.exp(contrib / temp) - 1 : 0;
	}

	/**
	 * @param x
	 * @return x * log(x)
	 */
	protected final static double xlogx(double x) {
		assert x >= 0;
		return x == 0 ? 0 : x * Math.log(x);
	}

	/**
	 * Reduce all the expressions contained in the ECT to the AND node, i.e.,
	 * all the appearances of these expressions are substituted with the AND
	 * node. Remove the stored pairs that are invalidated due to the reduction,
	 * e.g., "xa" in "xab" is invalidated by substituting "ab" to the new AND.
	 * <p>
	 * There is a special case, that two resulting AND nodes are neighboring in
	 * a sentence. In this case, for each of the two AND nodes, one of its
	 * context is invalidated and a new context is added into the ECT (e.g.,
	 * removing "xN" and adding "NN"). For simplicity, when adding such a new
	 * context for an expression, we always use the null expression instead.
	 * 
	 * @param ect
	 *            the ECT
	 * @param n
	 *            the AND node
	 */
	protected void doReduction(ExpressionContextMatrix ect, AndNode n) {
		HashSet<Element> cnN = new HashSet<Element>();
		HashSet<Element> cNn = new HashSet<Element>();

		Expression[] exps = ect.rows.keySet().toArray(new Expression[0]);
		for (Expression exp : exps) {
			Map<Context, Collection<Element>> row = ect.rows.get(exp);
			if (row == null)
				continue;
			Context[] keys = row.keySet().toArray(new Context[0]);
			for (Context key : keys) {
				// A pair may appear twice in ECT, one for left_neighbor
				// and one for right_neighbor. So we only do one of them.
				if (key.contextType == Context.LEFT_NEIGHBOR) {
					Collection<Element> app = row.get(key);
					if (app == null)
						continue;
					Element[] elems = app.toArray(new Element[0]);
					for (Element elem : elems) {
						if (!app.contains(elem))
							continue;

						// 1. invalidate two neighboring pairs unless
						// finding the "NN" special case
						if (elem.prev != null) {
							if (elem.prev.sym != null) {
								ssm.delAppearance(elem.prev);

								/*
								 * It's also possible that the neighboring pair
								 * is also in the bi-cluster. It's not very
								 * likely, so we don't check which of the two
								 * conflicting pairs is better, and simply
								 * reserve the first encountered one and delete
								 * the other. If this does occur, we also have
								 * to update the AND/OR nodes that are just
								 * created.
								 */
								/*
								 * We can't use ect.delAppearance() here because
								 * in ect, an expression's context may have been
								 * changed due to reduction, making ect no
								 * longer a valid E-C matrix.
								 */
								boolean found = false;
								Expression pair = new Expression(elem.prev.sym,
										elem.sym);
								Map<Context, Collection<Sentence.Element>> r = ect.rows
										.get(pair);
								if (r != null) {
									// TODO v3 with longer context, the
									// following may be wrong
									Context cont = new Context(elem.next.sym,
											Context.RIGHT_NEIGHBOR);
									Collection<Sentence.Element> c = r
											.get(cont);
									if (c != null)
										if (c.remove(elem.prev)) {
											found = true;
											if (c.size() == 0)
												ect.delCell(pair, cont);

											boolean context_lost = false;
											if (elem.prev.prev != null)
												if (elem.prev.prev.sym == null) {
													// special case: context
													// changed due to previous
													// reduction
													context_lost = true;
												}
											if (context_lost) {
												for (Map.Entry<Context, Collection<Sentence.Element>> e : r
														.entrySet())
													if (e.getKey().contextType == Context.LEFT_NEIGHBOR)
														if (e
																.getValue()
																.remove(
																		elem.prev)) {
															if (e.getValue()
																	.isEmpty())
																ect
																		.delCell(
																				pair,
																				e
																						.getKey());
															break;
														}
											} else {
												Symbol s = elem.prev.prev != null ? elem.prev.prev.sym
														: null;
												cont = new Context(s,
														Context.LEFT_NEIGHBOR);
												c = r.get(cont);
												assert c.contains(elem.prev);
												c.remove(elem.prev);
												if (c.size() == 0)
													ect.delCell(pair, cont);
											}
										}
								}
								if (found) {
									n.left.decreaseChildWeight(elem.prev.sym);
									n.right.decreaseChildWeight(elem.sym);
									n.appearanceNum--;
								}
							} else { // the "NN" special case
								cNn.add(elem.prev);
								cnN.add(elem);

								// Two contexts of the two "N" stored in ECT
								// have to be invalidated.
								Context cont1 = new Context(elem.sym,
										Context.RIGHT_NEIGHBOR);
								for (Map.Entry<Expression, Collection<Element>> e : ect.cols
										.get(cont1).entrySet()) {
									if (e.getValue().remove(elem.prev)) {
										Expression exp1 = e.getKey();
										Expression exp2 = new Expression(
												elem.sym, elem.next.sym);
										Context cont2 = new Context(exp1.s2,
												Context.LEFT_NEIGHBOR);

										if (e.getValue().isEmpty())
											ect.delCell(exp1, cont1);
										Collection<Element> apps = ect.getCell(
												exp2, cont2);
										assert apps.contains(elem);
										apps.remove(elem);
										if (apps.isEmpty())
											ect.delCell(exp2, cont2);

										if (DEBUG_MODE)
											cont1 = null;
										break;
									}
								}
								if (DEBUG_MODE)
									// assert that the removal must happen
									assert cont1 == null;
							}
						}
						if (elem.next.next != null) {
							if (elem.next.next.sym != null) {
								ssm.delAppearance(elem.next);

								boolean found = false;
								Expression pair = new Expression(elem.next.sym,
										elem.next.next.sym);
								Map<Context, Collection<Sentence.Element>> r = ect.rows
										.get(pair);
								if (r != null) {
									Context cont = new Context(elem.sym,
											Context.LEFT_NEIGHBOR);
									Collection<Sentence.Element> c = r
											.get(cont);
									if (c != null)
										if (c.remove(elem.next)) {
											found = true;
											if (c.size() == 0)
												ect.delCell(pair, cont);

											boolean context_lost = false;
											if (elem.next.next.next != null)
												if (elem.next.next.next.sym == null) {
													// special case: context
													// changed due to previous
													// reduction
													context_lost = true;
												}
											if (context_lost) {
												for (Map.Entry<Context, Collection<Sentence.Element>> e : r
														.entrySet())
													if (e.getKey().contextType == Context.RIGHT_NEIGHBOR)
														if (e
																.getValue()
																.remove(
																		elem.next)) {
															if (e.getValue()
																	.isEmpty())
																ect
																		.delCell(
																				pair,
																				e
																						.getKey());
															break;
														}
											} else {
												Symbol s = elem.next.next.next != null ? elem.next.next.next.sym
														: null;
												cont = new Context(s,
														Context.RIGHT_NEIGHBOR);
												c = r.get(cont);
												assert c.contains(elem.next);
												c.remove(elem.next);
												if (c.size() == 0)
													ect.delCell(pair, cont);
											}
										}
								}
								if (found) {
									n.left.decreaseChildWeight(elem.next.sym);
									n.right
											.decreaseChildWeight(elem.next.next.sym);
									n.appearanceNum--;
								}
							} else { // the "NN" special case
								cNn.add(elem);
								cnN.add(elem.next.next);

								// Two contexts of the two "N" stored in ECT
								// have to be invalidated.
								Context cont1 = new Context(elem.next.sym,
										Context.LEFT_NEIGHBOR);
								for (Map.Entry<Expression, Collection<Element>> e : ect.cols
										.get(cont1).entrySet()) {
									if (e.getValue().remove(elem.next.next)) {
										Expression exp1 = e.getKey();
										Expression exp2 = new Expression(
												elem.sym, elem.next.sym);
										Context cont2 = new Context(exp1.s1,
												Context.RIGHT_NEIGHBOR);

										if (e.getValue().isEmpty())
											ect.delCell(exp1, cont1);
										Collection<Element> apps = ect.getCell(
												exp2, cont2);
										assert apps.contains(elem);
										apps.remove(elem);
										if (apps.isEmpty())
											ect.delCell(exp2, cont2);

										if (DEBUG_MODE)
											cont1 = null;
										break;
									}
								}
								if (DEBUG_MODE)
									// assert that the removal must happen
									assert cont1 == null;
							}
						}

						// 2. substitute the pair to the new AND
						elem.sym = null; /*
											 * assign null temporarily, because
											 * if assign n, then the NN
											 * special-case judgment may be
											 * interfered, because we don't know
											 * whether a N is newly reduced or
											 * exists before the reduction.
											 */
						Element rn = elem.next.next;
						elem.next = rn;
						if (rn != null)
							rn.prev = elem;
					}
				}
			}
		}

		// Store cNN into ect by two special columns
		if (!cnN.isEmpty()) {
			Map<Expression, Collection<Element>> col;
			Context lc = new Context(n, Context.LEFT_NEIGHBOR);
			if (ect.cols.containsKey(lc))
				col = ect.cols.get(lc);
			else {
				col = new HashMap<Expression, Collection<Element>>();
				ect.cols.put(lc, col);
			}
			col.put(null, cnN);

			Context rc = new Context(n, Context.RIGHT_NEIGHBOR);
			if (ect.cols.containsKey(rc))
				col = ect.cols.get(rc);
			else {
				col = new HashMap<Expression, Collection<Element>>();
				ect.cols.put(rc, col);
			}
			col.put(null, cNn);
		}

		// Substitute null with n (see the above comment)
		for (Map.Entry<Context, Map<Expression, Collection<Element>>> e : ect.cols
				.entrySet()) {
			if (e.getKey().contextType == Context.LEFT_NEIGHBOR) {
				for (Collection<Element> app : e.getValue().values()) {
					for (Element elem : app) {
						assert elem.sym == null;
						elem.sym = n;
					}
				}
			}
		}

		// Update ssm cell context
		for (Map<Expression, Collection<Sentence.Element>> col : ect.cols
				.values())
			for (Collection<Sentence.Element> apps : col.values())
				for (Element elem : apps)
					updateSSMCellAfterReducing(elem);
	}

	/*------------------------------------------------*/
	/*------------- Algorithm Parameters -------------*/
	/*------------------------------------------------*/
	/**
	 * The type of ConstituentFilter to be used for constituent judgment.
	 * 
	 * @see ConstituentFilter
	 */
	public static int constituentFilterType = ConstituentFilter.NULL_FILTER;

	/**
	 * The threshold of the normalized residue for bi-cluster multiplicative
	 * coherence.
	 */
	// protected static double delta = 1; //for mode 1
	// public static double delta = 5; // for mode 2
	public static double delta = 0.5; // for v2

	/**
	 * The threshold of the normalized residue for expression-context matrix
	 * multiplicative coherence.
	 */
	// protected static double delta2 = 0.2; //for mode 1
	public static double delta2 = 0.5; // for mode 2

	/**
	 * The parameter of the grammar prior probability. In the basic form, this
	 * should be the number of bits needed to represent a symbol in the grammar.
	 * We may simply use the logarithm of the number of different symbols used
	 * in the grammar. Moreover, we may use this parameter to adjust the
	 * importance of the prior in learning, by multiplying it with a
	 * non-negative factor. The larger it is, the more important the prior is
	 * (and usually the more generalized the learning is).
	 */
	public static double alpha = 0.2; // TODO relate to the corpus size

	/*
	 * Some observations of delta, delta2, alpha:
	 * 
	 * When alpha is too large, the bi-cluster found by the 1st hill-climbing
	 * may often not be able to pass the delta threshold (unless delta is also
	 * very large), leading to slow or failed bi-clustering.
	 * 
	 * When delta exceeds a certain point, it doesn't have much effect, because
	 * after that point it simply passes almost all the bi-clusters (remember
	 * that those bi-clusters are often good mcbc unless alpha is large).
	 */

	/**
	 * The maximally allowed ratio of the real value to the predicted value of a
	 * cell in a multiplicative coherent bi-cluster. If the ratio is higher than
	 * this, we think that cell may involve overlapping bi-clusters.
	 */
	public static double mcbcMaxRatioToPrediction = 3;

	/**
	 * In {@link #shcContrib2Weight(double)}, we compute the exponential of the
	 * contribution as the weight. This parameter controls the steepness of the
	 * exponential curve for the first hill-climbing in biClustering().
	 */
	public static double T1 = 0.01;

	/**
	 * In {@link #shcContrib2Weight(double)}, we compute the exponential of the
	 * contribution as the weight. This parameter controls the steepness of the
	 * exponential curve for the second hill-climbing in biClustering().
	 */
	// public static double T2 = 0.01;
	/**
	 * For a one-cell bi-cluster, if its cell value is larger than or equal to
	 * this threshold, then it's accepted and extracted.
	 * 
	 * TODO Enhancement: Ideally, this threshold should be related to the corpus
	 * size (sentence#) and maybe some other factors.
	 */
	public static int oneCellBiClusterThreshold = 4;

	/**
	 * The number of bi-clustering attempted before giving up. We are using
	 * local search to find a valid bi-cluster, so we have to try reasonably
	 * many times before claiming that there (probably) doesn't exist any valid
	 * bi-cluster.
	 */
	public static int biClusteringAttemptNum = 10;

	/**
	 * The maximal recursive depth allowed in inferContext(). The large it is,
	 * the more accurate the inferred contexts are, but the slower the program
	 * becomes.
	 */
	public static int icMaxRecursiveDepth = 10;

	/**
	 * 1: by sum; 2: by size.
	 */
	public static final int NORMALIZATION_MODE = 2;

	/**
	 * The number of bi-clusters we try to find in bestBiClustering(), among
	 * which the one with the highest posterior gain is returned. In principle,
	 * the larger this number is, the better quality the learned grammar might
	 * have, and the slower the algorithm is. This number should be less than
	 * the expected number of bi-clusters in the symbol-symbol matrix: being
	 * larger than that would simply slow down the algorithm without any further
	 * performance gain.
	 */
	public static int numSimultaneousBiClustering = 10;

	/*------------------------------------------------*/
	/*------ Implementation Specific Parameters ------*/
	/*------------------------------------------------*/
	/**
	 * For {@link SymbolSymbolMatrix}, the row/col set is a HashMap, the
	 * initial capacity of which is the number of terms in the corpus times this
	 * value. In general, the larger it is, the less time and more space the
	 * program takes.
	 * <p>
	 * Note that it is said that the initial capacity should be twice the size
	 * to which you expect the set to grow. So this value should be 2 times the
	 * ratio of the expected maximal SymbolSymbolMatrix size to the term number.
	 */
	protected static final int hashMapInitCapMult = 3;

	/**
	 * The estimated maximal size of the row/column set.
	 * <p>
	 * This is used in {@link BiCluster} to determine the initial capacity of
	 * the HashSet that stores the row/col set. It depends on the property of
	 * the grammar to be learned.
	 */
	protected static final int biClusterMaxSize = 30;

	/**
	 * Used to control whether to print out debug info. By making it final, JVM
	 * should be able to ignore the codes at compiling time when it's false.
	 */
	protected static final boolean DEBUG_MODE = false;

	/*------------------------------------------------*/
	/*----------- Variables for Debugging ------------*/
	/*------------------------------------------------*/
	/* for the learning history */
	/**
	 * record the printout of all the ssm/bc/bc.ect in the learning process
	 */
	protected ArrayList<String> debug_learning_history = new ArrayList<String>();
	/* for BiClustering */
	/**
	 * number of bi-clusters resulting from the 1st hill-climbing
	 */
	protected static int debug_bc_nHC1 = 0;
	/**
	 * number of multi-row multi-column bi-clusters resulting from the 1st
	 * hill-climbing
	 */
	// protected static int debug_bc_nMrMcBC = 0;
	/**
	 * number of bi-clusters that fail the 1st criterion, i.e., two
	 * multiplicative coherence
	 */
	protected static int debug_bc_nFailedCheck1 = 0;
	/**
	 * number of single-cell bi-clusters resulting from the 1st hill-climbing
	 */
	// protected static int debug_bc_nScBC = 0;
	/**
	 * number of single-cell bi-clusters after the 2nd hill-climbing
	 */
	protected static int debug_bc_nScBC2 = 0;
	/**
	 * number of single-cell bi-clusters that fail the 3rd criterion, i.e.,
	 * single-cell bi-cluster threshold
	 */
	protected static int debug_bc_nFailedCheck3 = 0;
	/* for Overlap Filtering */
	/**
	 * When doing bi-clustering, the total number of appearances in bi-clusters
	 * before overlap-filtering
	 */
	protected static int debug_of_bc_nBeforeOF = 0;
	/**
	 * When doing bi-clustering, the total number of appearances in bi-clusters
	 * after overlap-filtering
	 */
	protected static int debug_of_bc_nAfterOF = 0;
	/**
	 * When merging new ANDs to existing ORs, the total number of appearances in
	 * to-be-merged part before overlap-filtering
	 */
	protected static int debug_of_mg_nBeforeOF = 0;
	/**
	 * When merging new ANDs to existing ORs, the total number of appearances in
	 * to-be-merged part after overlap-filtering
	 */
	protected static int debug_of_mg_nAfterOF = 0;
	/* for merging new ANDs to existing ORs */
	/**
	 * number of proposed merging candidates
	 */
	protected static int debug_mg_nCandidate = 0;
	/**
	 * number of merging candidates that pass the 1st criterion, i.e.,
	 * bi-cluster multiplicative coherence
	 */
	protected static int debug_mg_nPastCheck1 = 0;
	/**
	 * number of merging candidates that pass the 2nd criterion, i.e., ECT
	 * multiplicative coherence. This is also the number of actually merged
	 * candidates.
	 */
	protected static int debug_mg_nPastCheck2 = 0;
}

/*
 * TODO When running large scale tests: 1) since we use assertions in the code,
 * don't use the -ea (enable assertion) VM argument; 2) set DEBUG_MODE to false
 * to disable debugging codes; 3) use "-server" VM argument to use the Java
 * Server VM (faster, but more space and startup time).
 * 
 * We may also use greedy hill-climbing instead of stochastic hill-climbing in
 * bi-clustering to speed up the program. It is even questionable whether the
 * stochastic hc is better than the greedy one, as experiments show both
 * positive and negative results for different data.
 */
