package gi;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * The elements of Matrix are of type {@code Collection<Sentence.Element>},
 * which stores all the appearances of the corresponding pair (pointing to the
 * 1st element by default) and thus also the appearance number.
 * <p>
 * For {@link SymbolSymbolMatrix}, we use HashSet, because we want to traverse,
 * add, remove its elements, and its initial size is known at creation (except
 * for the initial SymbolSymbolMatrix, which may be solved by an initial
 * indexing of the corpus).
 * <p>
 * For {@link ExpressionContextMatrix}, we use ArrayList, because we want to
 * traverse and add its elements, and its initial size is unknown.
 * 
 * @author Kewei Tu
 * 
 * @param <RH>
 *            the type of row headers
 * @param <CH>
 *            the type of column headers
 */
public abstract class Matrix<RH, CH, CELL> {
	protected Map<RH, Map<CH, CELL>> rows;
	protected Map<CH, Map<RH, CELL>> cols;

	public Matrix() {
		rows = new HashMap<RH, Map<CH, CELL>>();
		cols = new HashMap<CH, Map<RH, CELL>>();
	}

	protected void addRow(RH s) {
		rows.put(s, new HashMap<CH, CELL>());
	}

	protected void addCol(CH s) {
		cols.put(s, new HashMap<RH, CELL>());
	}

	public void delRow(RH s) {
		Map<CH, CELL> m = rows.remove(s);
		if (m != null)
			for (CH c : m.keySet()) {
				Map<RH, CELL> m2 = cols.get(c);
				m2.remove(s);
				if (m2.isEmpty())
					cols.remove(c);
			}
	}

	public void delCol(CH s) {
		Map<RH, CELL> m = cols.remove(s);
		if (m != null)
			for (RH r : m.keySet()) {
				Map<CH, CELL> m2 = rows.get(r);
				m2.remove(s);
				if (m2.isEmpty())
					rows.remove(r);
			}
	}

	public void delCell(RH r, CH c) {
		Map<CH, CELL> row = rows.get(r);
		Map<RH, CELL> col = cols.get(c);
		if (row != null && col != null) {
			row.remove(c);
			col.remove(r);
			if (row.isEmpty())
				delRow(r);
			if (col.isEmpty())
				delCol(c);
		}
	}

	public CELL getCell(RH row, CH col) {
		Map<CH, CELL> r = rows.get(row);
		if (r != null)
			return r.get(col);
		return null;
	}

	public void addCell(RH r, CH c, CELL apps) {
		if (!rows.containsKey(r))
			addRow(r);
		if (!cols.containsKey(c))
			addCol(c);
		rows.get(r).put(c, apps);
		cols.get(c).put(r, apps);
	}

	public boolean isEmpty() {
		return rows.isEmpty();
	}

	/*
	 * For debugging: display the matrix if small
	 */
	@Override
	public String toString() {
		String re = "size: " + rows.size() + "*" + cols.size() + "\n";

		final int sizelimit = 50;
		ArrayList<CH> cs = new ArrayList<CH>(sizelimit);
		int count = 0;
		for (CH c : cols.keySet()) {
			cs.add(c);
			count++;
			if (count == sizelimit)
				break;
		}
		for (Object ch : cs)
			re += "\t" + ch;
		if (cols.size() > sizelimit)
			re += "\t(" + (cols.size() - sizelimit) + " more cols omitted)";
		count = 0;
		for (Map.Entry<RH, Map<CH, CELL>> r : rows.entrySet()) {
			re += "\n" + r.getKey();
			for (Object ch : cs) {
				re += "\t";
				CELL c = r.getValue().get(ch);
				if (c != null)
					re += cellToString(c);
			}
			count++;
			if (count == sizelimit)
				break;
		}
		if (rows.size() > sizelimit)
			re += "\n(" + (rows.size() - sizelimit) + " more rows omitted)";

		return re;
	}

	protected abstract String cellToString(CELL c);

	public void printToFile(String filename) {
		try {
			FileWriter fw = new FileWriter(filename);

			fw.write("size: " + rows.size() + "*" + cols.size() + "\n");
			ArrayList<CH> cs = new ArrayList<CH>(cols.keySet());
			for (Object ch : cs)
				fw.write("\t" + ch);
			for (Map.Entry<RH, Map<CH, CELL>> r : rows.entrySet()) {
				fw.write("\n" + r.getKey());
				for (Object ch : cs) {
					fw.write("\t");
					CELL c = r.getValue().get(ch);
					if (c != null)
						fw.write(cellToString(c));
				}
			}

			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
