package gi;

import gi.ExpressionContextMatrix.Context;
import gi.ExpressionContextMatrix.Expression;
import gi.Sentence.Element;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

/**
 * The symbol-symbol matrix used in the grammar learning algorithm. Each cell of
 * the matrix represents a pair of symbols and contains a set of pointers each
 * pointing to an appearance of the pair (at the first symbol of the pair).
 * 
 * @author Kewei Tu
 * 
 */
public class SymbolSymbolMatrix extends
		Matrix<Symbol, Symbol, SymbolSymbolMatrix.SymbolSymbolCell> {

	static public class SymbolSymbolCell {
		public int num = 0;
		public HashMap<ExpressionContextMatrix.Context, Collection<Element>> ecv = new HashMap<ExpressionContextMatrix.Context, Collection<Element>>();

		public boolean isConsistent() {
			int count = 0;
			for (Collection<Element> c : ecv.values()) {
				count += c.size();
			}
			return count / 2 == num; // TODO v3 del /2
		}

		public boolean remove(Element app) {
			boolean foundone = false;
			for (Iterator<Collection<Element>> iter = ecv.values().iterator(); iter
					.hasNext();) {
				Collection<Element> c = (Collection<Element>) iter.next();
				if (c.remove(app)) {
					if (c.isEmpty())
						iter.remove();

					if (foundone) {
						num--;
						return true;
					} else
						foundone = true;
					// TODO v3 no need "foundone"
				}
			}
			return false;
		}

		public void removeAll(Map<Context, Collection<Element>> ecv2) {
			for (Map.Entry<Context, Collection<Element>> e : ecv2.entrySet()) {
				Collection<Element> c = ecv.get(e.getKey());
				if (c != null)
					if (c.removeAll(e.getValue()))
						if (c.isEmpty())
							ecv.remove(e.getKey());
			}
			num = 0;
			for (Collection<Element> c : ecv.values()) {
				num += c.size();
			}
			num /= 2; // TODO v3 del /2
		}

		public boolean isEmpty() {
			return num == 0;
		}

		public void add(Element elem) { // TODO v3 change
			num++;

			// left
			Symbol s = elem.prev != null ? elem.prev.sym : null;
			Context cont = new Context(s, Context.LEFT_NEIGHBOR);
			Collection<Element> c = ecv.get(cont);
			if (c == null) {
				c = new HashSet<Element>();
				ecv.put(cont, c);
			}
			c.add(elem);

			// right
			s = elem.next.next != null ? elem.next.next.sym : null;
			cont = new Context(s, Context.RIGHT_NEIGHBOR);
			c = ecv.get(cont);
			if (c == null) {
				c = new HashSet<Element>();
				ecv.put(cont, c);
			}
			c.add(elem);
		}

		public void addAll(Collection<Element> app) {
			for (Element element : app)
				add(element);
		}

		public int size() {
			return num;
		}

		@Override
		public String toString() {
			String re = "size: 1*" + ecv.size() + "\n";
			if (ecv.size() <= 50) {
				Object[] cs = ecv.keySet().toArray();
				for (Object ch : cs)
					re += ch + "\t";
				re += "\n";
				for (Object ch : cs) {
					Collection<Element> c = ecv.get(ch);
					if (c != null)
						re += c.size();
					re += "\t";
				}
			}
			return re;
		}
	}

	/**
	 * For fast random selection
	 */
	// protected ArrayList<Symbol> rowHeaders, colHeaders;
	protected Collection<Expression> cells;

	public SymbolSymbolMatrix() {
		super();
		// rowHeaders = new ArrayList<Symbol>();
		// colHeaders = new ArrayList<Symbol>();
		cells = new HashSet<Expression>();
	}

	/**
	 * Construct an empty symbol-symbol matrix, the size of which is determined
	 * by {@code termNum}.
	 * 
	 * @param termNum
	 *            the number of terms in the corpus based on which this matrix
	 *            is initialized.
	 */
	public SymbolSymbolMatrix(int termNum) {
		rows = new HashMap<Symbol, Map<Symbol, SymbolSymbolCell>>(termNum
				* GrammarLearning.hashMapInitCapMult);
		cols = new HashMap<Symbol, Map<Symbol, SymbolSymbolCell>>(termNum
				* GrammarLearning.hashMapInitCapMult);
		// rowHeaders = new ArrayList<Symbol>(termNum);
		// colHeaders = new ArrayList<Symbol>(termNum);
		cells = new HashSet<Expression>(termNum
				* GrammarLearning.hashMapInitCapMult);
	}

	/**
	 * Construct a symbol-symbol matrix based on a corpus.
	 * 
	 * @param c
	 *            the corpus based on which this matrix is initialized.
	 */
	public SymbolSymbolMatrix(Corpus c) {
		this(c.terms.length);
		for (Sentence sen : c.sentences) {
			Sentence.Element e1 = sen.firstElem;
			if (e1 != null) {
				Sentence.Element e2 = e1.next;
				while (e2 != null) {
					if (!rows.containsKey(e1.sym))
						addRow(e1.sym);
					if (!cols.containsKey(e2.sym))
						addCol(e2.sym);
					SymbolSymbolCell appearances = rows.get(e1.sym).get(e2.sym);
					if (appearances == null) {
						appearances = new SymbolSymbolCell();
						rows.get(e1.sym).put(e2.sym, appearances);
						cols.get(e2.sym).put(e1.sym, appearances);
						cells.add(new Expression(e1.sym, e2.sym));
					}
					appearances.add(e1);

					e1 = e2;
					e2 = e2.next;
				}
			}
		}
	}

	// @Override
	// public void addCol(Symbol s) {
	// cols.put(s, new HashMap<Symbol, Collection<Sentence.Element>>());
	// colHeaders.add((Symbol) s);
	// }
	//
	// @Override
	// public void addRow(Symbol s) {
	// rows.put(s, new HashMap<Symbol, Collection<Sentence.Element>>());
	// rowHeaders.add((Symbol) s);
	// }
	//
	// @Override
	// public void delCol(Symbol s) {
	// super.delCol(s);
	// colHeaders.remove(s);
	// }
	//
	// @Override
	// public void delRow(Symbol s) {
	// super.delRow(s);
	// rowHeaders.remove(s);
	// }

	public void delAppearance(Sentence.Element app) {
		Symbol row = app.sym;
		Symbol col = app.next.sym;
		Map<Symbol, SymbolSymbolCell> r = rows.get(row);
		if (r != null) {
			SymbolSymbolCell apps = r.get(col);
			if (apps != null)
				if (apps.remove(app))
					if (apps.isEmpty())
						delCell(row, col);
		}
	}

	public void addRow(Symbol s, Map<Symbol, SymbolSymbolCell> newRow) {
		rows.put(s, newRow);
		// rowHeaders.add((Symbol) s);
		for (Map.Entry<Symbol, SymbolSymbolCell> e : newRow.entrySet()) {
			Symbol s2 = e.getKey();
			if (!cols.containsKey(s2))
				addCol(s2);
			cols.get(s2).put(s, e.getValue());
			cells.add(new Expression(s, s2));
		}
	}

	public void addCol(Symbol s, Map<Symbol, SymbolSymbolCell> newCol) {
		cols.put(s, newCol);
		// colHeaders.add((Symbol) s);
		for (Map.Entry<Symbol, SymbolSymbolCell> e : newCol.entrySet()) {
			Symbol s2 = e.getKey();
			if (!rows.containsKey(s2))
				addRow(s2);
			rows.get(s2).put(s, e.getValue());
			cells.add(new Expression(s2, s));
		}
	}

	@Override
	public void addCell(Symbol r, Symbol c, SymbolSymbolCell apps) {
		super.addCell(r, c, apps);
		cells.add(new Expression(r, c));
	}

	@Override
	public void delCell(Symbol s1, Symbol s2) {
		SymbolSymbolCell apps = getCell(s1, s2);
		if (apps != null)
			cells.remove(new Expression(s1, s2));
		super.delCell(s1, s2);
	}

	@Override
	public void delCol(Symbol s) {
		Map<Symbol, SymbolSymbolCell> m = cols.remove(s);
		if (m != null)
			for (Symbol r : m.keySet()) {
				Map<Symbol, SymbolSymbolCell> m2 = rows.get(r);
				m2.remove(s);
				cells.remove(new Expression(r, s));
				if (m2.isEmpty())
					rows.remove(r);
			}
	}

	@Override
	public void delRow(Symbol s) {
		Map<Symbol, SymbolSymbolCell> m = rows.remove(s);
		if (m != null)
			for (Symbol c : m.keySet()) {
				Map<Symbol, SymbolSymbolCell> m2 = cols.get(c);
				m2.remove(s);
				cells.remove(new Expression(s, c));
				if (m2.isEmpty())
					cols.remove(c);
			}
	}

	/**
	 * For debugging. Check if the ssm is consistent, i.e. appearances stored in
	 * each cell are consistent with the cell position.
	 * 
	 * @return
	 */
	protected boolean isConsistent() {
		for (Map.Entry<Symbol, Map<Symbol, SymbolSymbolCell>> e1 : rows
				.entrySet()) {
			for (Map.Entry<Symbol, SymbolSymbolCell> e2 : e1.getValue()
					.entrySet()) {
				SymbolSymbolCell c = e2.getValue();
				assert !c.isEmpty();
				assert c.isConsistent();
				for (Collection<Element> d : c.ecv.values())
					for (Sentence.Element app : d) {
						if (app.next == null)
							return false;
						if (app.sym != e1.getKey()
								|| app.next.sym != e2.getKey())
							return false;
					}
			}
		}
		return true;
	}

	protected void printInconsistency() {
		for (Map.Entry<Symbol, Map<Symbol, SymbolSymbolCell>> e1 : rows
				.entrySet()) {
			for (Map.Entry<Symbol, SymbolSymbolCell> e2 : e1.getValue()
					.entrySet()) {
				for (Collection<Element> d : e2.getValue().ecv.values())
					for (Sentence.Element app : d) {
						if (app.next == null)
							System.out.println(e1.getKey() + "," + e2.getKey()
									+ ":" + app);
						else if (app.sym != e1.getKey()
								|| app.next.sym != e2.getKey())
							System.out.println(e1.getKey() + "," + e2.getKey()
									+ ":" + app);
					}
			}
		}
	}

	@Override
	protected String cellToString(SymbolSymbolCell c) {
		return Integer.toString(c.num);
	}
}
