package gi.util;

import gi.AndNode;
import gi.Grammar;
import gi.OrNode;
import gi.Sentence;
import gi.Symbol;
import gi.Term;
import gi.Sentence.Element;

import java.util.HashMap;
import java.util.Map;

/**
 * A CYK Parser for CNF-PCFG.
 * 
 * @author Kewei Tu
 * 
 */
public class CYKParser {
	protected Grammar grammar = null;
	protected Sentence sentence = null;

	/**
	 * current sentence length
	 */
	protected int length = 0;

	protected class ParsingNode {
		protected Symbol s;
		protected double prob;
		protected Symbol left, right;
		protected int k;

		protected ParsingNode(Symbol s) {
			this.s = s;
			prob = 1;
			left = right = null;
		}

		public ParsingNode(Symbol s, double prob, Symbol left, Symbol right,
				int k) {
			this.s = s;
			this.prob = prob;
			this.left = left;
			this.right = right;
			this.k = k;
		}
	}

	/**
	 * The CYK parsing table. For table[i][j], it stores symbols that can
	 * generate the substring from the j-th term to (j+i)-th term in the
	 * sentence.
	 * 
	 * Java doesn't support generic array, so "SuppressWarnings" has to be used.
	 * 
	 * We reuse the table for different parsings in order to save the cost of
	 * recreating the table. So usually only a part of the table is actually
	 * used by the current parsing, which can be found out using the current
	 * sentence length.
	 */
	@SuppressWarnings("unchecked")
	protected Map<Symbol, ParsingNode>[][] table = new HashMap[0][0];

	public void setGrammar(Grammar grammar) {
		this.grammar = grammar;
	}

	/**
	 * Parse a sentence with the current grammar, and return whether it succeeds
	 * or not. This method tries to find the best parsing (i.e. the parsing with
	 * the highest probability), so it doesn't save every possible parsing in
	 * the CYK table. The resulting CYK table is stored, which can later be used
	 * for further queries of the parsing.
	 * 
	 * @param sen
	 * @return
	 */
	@SuppressWarnings("unchecked")
	// TODO Enhancement: accelerate by (re-)using ParsingNode, two mode choice,
	// (symbol*symbol)-rule map, etc.
	public boolean parse(Sentence sen) {
		if (grammar == null | sen == null)
			return false;

		sentence = sen;
		length = sen.length();

		if (length == 0)
			return false;
		// Currently our grammar doesn't support empty string.

		// Initialize the table
		// if the table is not big enough for the current sentence, expand it.
		if (table.length < length)
			if (table.length * 2 <= length)
				table = new HashMap[length][length];
			else
				table = new HashMap[table.length * 2][table.length * 2];
		int i = 0;
		for (Element e = sen.firstElem; e != null; e = e.next) {
			if (!(e.sym instanceof Term))
				return false; // shouldn't happen.
			Term t = grammar.getTerminal(((Term) e.sym).obj);
			if (t == null)
				return false; // there's a new term, so can't be parsed.
			if (table[0][i] == null)
				table[0][i] = new HashMap<Symbol, ParsingNode>(1, 1);
			else
				table[0][i].clear();
			table[0][i].put(t, new ParsingNode(t));
			i++;
		}

		// The main loop
		for (i = 1; i < length; i++) {
			for (int j = 0; j < length - i; j++) {
				if (table[i][j] == null)
					table[i][j] = new HashMap<Symbol, ParsingNode>();
				else
					table[i][j].clear();
				for (int k = 0; k < i; k++) {
					Map<Symbol, ParsingNode> left = table[k][j];
					Map<Symbol, ParsingNode> right = table[i - k - 1][j + k + 1];
					if (left.isEmpty() || right.isEmpty())
						continue;
					for (ParsingNode ln : left.values()) {
						for (OrNode lor : ln.s.parents) {
							for (AndNode land : lor.parents) {
								if (land.left != lor)
									continue;
								for (ParsingNode rn : right.values()) {
									for (OrNode ror : rn.s.parents) {
										if (land.right == ror) {
											// double div = land.appearanceNum;
											double div = 0;
											for (double[] w : lor.weights) {
												div += w[0];
											}
											double lp = lor.weights
													.get(lor.children
															.indexOf(ln.s))[0]
													/ div;
											div = 0;
											for (double[] w : ror.weights) {
												div += w[0];
											}
											double rp = ror.weights
													.get(ror.children
															.indexOf(rn.s))[0]
													/ div;
											assert lp != 0 && rp != 0;
											double p = ln.prob * rn.prob * lp
													* rp;
											ParsingNode n = table[i][j]
													.get(land);
											if (n != null) {
												if (n.prob < p) {
													n.prob = p;
													n.left = ln.s;
													n.right = rn.s;
													n.k = k;
												}
											} else
												table[i][j].put(land,
														new ParsingNode(land,
																p, ln.s, rn.s,
																k));
										}
									}
								}
							}
						}
					}
				}
			}
		}

		// Return whether parsed or not
		Map<Symbol, ParsingNode> s = table[length - 1][0];
		if (s.isEmpty())
			return false;
		else
			for (Symbol n : s.keySet())
				if (grammar.startSymbol.children.contains(n))
					return true;
		return false;
	}

	/**
	 * @return a string of the bracketing of the sentence, according to the
	 *         EVALB format. All the nodes, including terminals and
	 *         nonterminals, are labeled "x". E.g., (x (x this) (x (x is) (x (x
	 *         an) (x example)))).
	 */
	public String getBestBracketingString() {
		Map<Symbol, ParsingNode> entry = table[length - 1][0];
		Symbol best = null;
		double max = 0;
		if (entry != null)
			for (ParsingNode p : entry.values())
				if (p.prob > max) {
					max = p.prob;
					best = p.s;
				}
		if (best != null)
			return getBestBracketingString2(length - 1, 0, best);
		else
			return "";
	}

	protected String getBestBracketingString2(int i, int j, Symbol s) {
		Map<Symbol, ParsingNode> entry = table[i][j];
		if (i == 0) { // terminal
			return "(x " + entry.keySet().iterator().next().toString() + ")";
		} else { // nonterminal
			ParsingNode pn = entry.get(s);
			return "(x "
					+ getBestBracketingString2(pn.k, j, pn.left)
					+ " "
					+ getBestBracketingString2(i - pn.k - 1, j + pn.k + 1,
							pn.right) + ")";
		}

	}

	/**
	 * @param template
	 * @return a string of partial bracketing of the sentence, i.e., a forest
	 *         instead of a tree, the roots of which are presented in the
	 *         template sentence. The format is as in getBestBracketingString().
	 */
	public String getPartialBracketingString(Sentence template) {
		Element elem = template.firstElem;
		// dp[i][j]: the best prob that the first i+1 roots cover index
		// 0~j(inclusive) in the original sentence
		double[][] dp = new double[template.length()][length];
		// dpi[i][j]: for the best prob in dp[i][j], which index the (i+1)-th
		// root should start from in the original sentence
		int[][] dpi = new int[template.length()][length]; // segmenting index
		for (int i = 0; i < length - 1; i++) {
			ParsingNode pn = table[i][0].get(elem.sym);
			if (pn != null) {
				dp[0][i] = pn.prob;
			}
		}
		for (int i = 1; i < dp.length; i++) {
			elem = elem.next;
			for (int j = 0; j < length - 1; j++) { // len-1 for it shouldn't
				// cover the whole sentence
				if (dp[i - 1][j] != 0) {
					double p = dp[i - 1][j];
					for (int k = 0; k < length - j - 1; k++) {
						ParsingNode pn = table[k][j + 1].get(elem.sym);
						if (pn != null) {
							if (dp[i][j + k + 1] < pn.prob * p) {
								dp[i][j + k + 1] = pn.prob * p;
								dpi[i][j + k + 1] = j + 1;
							}
						}
					}
				}
			}
		}
		if (dp[dp.length - 1][length - 1] != 0) {
			String re = "";
			int right = length - 1;
			for (int i = dp.length - 1; i >= 0; i--) {
				int left = dpi[i][right];
				re = getBestBracketingString2(right - left, left, elem.sym)
						+ " " + re;
				right = left - 1;
				elem = elem.prev;
			}
			return "(x " + re + ")";
		} else
			return "";
	}

	// TODO getBestParsingTree, getPartialParsingTree, etc.
}
